# -*- coding: utf-8 -*-
#----------------------------------------------------------
# OpenERP Web HTTP layer
#----------------------------------------------------------
import ast
import cgi
import contextlib
import functools
import getpass
import logging
import mimetypes
import os
import pprint
import random
import sys
import tempfile
import threading
import time
import traceback
import urlparse
import uuid
import xmlrpclib
import errno

import babel.core
import simplejson
import werkzeug.contrib.sessions
import werkzeug.datastructures
import werkzeug.exceptions
import werkzeug.utils
import werkzeug.wrappers
import werkzeug.wsgi

import openerp

import session

_logger = logging.getLogger(__name__)

#----------------------------------------------------------
# RequestHandler
#----------------------------------------------------------
class WebRequest(object):
    """ Parent class for all OpenERP Web request types, mostly deals with
    initialization and setup of the request object (the dispatching itself has
    to be handled by the subclasses)

    :param request: a wrapped werkzeug Request object
    :type request: :class:`werkzeug.wrappers.BaseRequest`

    .. attribute:: httprequest

        the original :class:`werkzeug.wrappers.Request` object provided to the
        request

    .. attribute:: httpsession

        a :class:`~collections.Mapping` holding the HTTP session data for the
        current http session

    .. attribute:: params

        :class:`~collections.Mapping` of request parameters, not generally
        useful as they're provided directly to the handler method as keyword
        arguments

    .. attribute:: session_id

        opaque identifier for the :class:`session.OpenERPSession` instance of
        the current request

    .. attribute:: session

        :class:`~session.OpenERPSession` instance for the current request

    .. attribute:: context

        :class:`~collections.Mapping` of context values for the current request

    .. attribute:: debug

        ``bool``, indicates whether the debug mode is active on the client
    """
    def __init__(self, request):
        self.httprequest = request
        self.httpresponse = None
        self.httpsession = request.session

    def init(self, params):
        self.params = dict(params)
        # OpenERP session setup
        self.session_id = self.params.pop("session_id", None) or uuid.uuid4().hex
        self.session = self.httpsession.get(self.session_id)
        if not self.session:
            self.session = session.OpenERPSession()
            self.httpsession[self.session_id] = self.session

        # set db/uid trackers - they're cleaned up at the WSGI
        # dispatching phase in openerp.service.wsgi_server.application
        if self.session._db:
            threading.current_thread().dbname = self.session._db
        if self.session._uid:
            threading.current_thread().uid = self.session._uid

        self.context = self.params.pop('context', {})
        self.debug = self.params.pop('debug', False) is not False
        # Determine self.lang
        lang = self.params.get('lang', None)
        if lang is None:
            lang = self.context.get('lang')
        if lang is None:
            lang = self.httprequest.cookies.get('lang')
        if lang is None:
            lang = self.httprequest.accept_languages.best
        if not lang:
            lang = 'en_US'
        # tranform 2 letters lang like 'en' into 5 letters like 'en_US'
        lang = babel.core.LOCALE_ALIASES.get(lang, lang)
        # we use _ as seprator where RFC2616 uses '-'
        self.lang = lang.replace('-', '_')

def reject_nonliteral(dct):
    if '__ref' in dct:
        raise ValueError(
            "Non literal contexts can not be sent to the server anymore (%r)" % (dct,))
    return dct

class JsonRequest(WebRequest):
    """ JSON-RPC2 over HTTP.

    Sucessful request::

      --> {"jsonrpc": "2.0",
           "method": "call",
           "params": {"session_id": "SID",
                      "context": {},
                      "arg1": "val1" },
           "id": null}

      <-- {"jsonrpc": "2.0",
           "result": { "res1": "val1" },
           "id": null}

    Request producing a error::

      --> {"jsonrpc": "2.0",
           "method": "call",
           "params": {"session_id": "SID",
                      "context": {},
                      "arg1": "val1" },
           "id": null}

      <-- {"jsonrpc": "2.0",
           "error": {"code": 1,
                     "message": "End user error message.",
                     "data": {"code": "codestring",
                              "debug": "traceback" } },
           "id": null}

    """
    def dispatch(self, method):
        """ Calls the method asked for by the JSON-RPC2 or JSONP request

        :param method: the method which received the request

        :returns: an utf8 encoded JSON-RPC2 or JSONP reply
        """
        args = self.httprequest.args
        jsonp = args.get('jsonp')
        requestf = None
        request = None
        request_id = args.get('id')

        if jsonp and self.httprequest.method == 'POST':
            # jsonp 2 steps step1 POST: save call
            self.init(args)
            self.session.jsonp_requests[request_id] = self.httprequest.form['r']
            headers=[('Content-Type', 'text/plain; charset=utf-8')]
            r = werkzeug.wrappers.Response(request_id, headers=headers)
            return r
        elif jsonp and args.get('r'):
            # jsonp method GET
            request = args.get('r')
        elif jsonp and request_id:
            # jsonp 2 steps step2 GET: run and return result
            self.init(args)
            request = self.session.jsonp_requests.pop(request_id, "")
        else:
            # regular jsonrpc2
            requestf = self.httprequest.stream

        response = {"jsonrpc": "2.0" }
        error = None
        try:
            # Read POST content or POST Form Data named "request"
            if requestf:
                self.jsonrequest = simplejson.load(requestf, object_hook=reject_nonliteral)
            else:
                self.jsonrequest = simplejson.loads(request, object_hook=reject_nonliteral)
            self.init(self.jsonrequest.get("params", {}))
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("--> %s.%s\n%s", method.im_class.__name__, method.__name__, pprint.pformat(self.jsonrequest))
            response['id'] = self.jsonrequest.get('id')
            response["result"] = method(self, **self.params)
        except session.AuthenticationError:
            error = {
                'code': 100,
                'message': "OpenERP Session Invalid",
                'data': {
                    'type': 'session_invalid',
                    'debug': traceback.format_exc()
                }
            }
        except xmlrpclib.Fault, e:
            error = {
                'code': 200,
                'message': "OpenERP Server Error",
                'data': {
                    'type': 'server_exception',
                    'fault_code': e.faultCode,
                    'debug': "Client %s\nServer %s" % (
                    "".join(traceback.format_exception("", None, sys.exc_traceback)), e.faultString)
                }
            }
        except Exception:
            logging.getLogger(__name__ + '.JSONRequest.dispatch').exception\
                ("An error occured while handling a json request")
            error = {
                'code': 300,
                'message': "OpenERP WebClient Error",
                'data': {
                    'type': 'client_exception',
                    'debug': "Client %s" % traceback.format_exc()
                }
            }
        if error:
            response["error"] = error

        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("<--\n%s", pprint.pformat(response))

        if jsonp:
            # If we use jsonp, that's mean we are called from another host
            # Some browser (IE and Safari) do no allow third party cookies
            # We need then to manage http sessions manually.
            response['httpsessionid'] = self.httpsession.sid
            mime = 'application/javascript'
            body = "%s(%s);" % (jsonp, simplejson.dumps(response),)
        else:
            mime = 'application/json'
            body = simplejson.dumps(response)

        r = werkzeug.wrappers.Response(body, headers=[('Content-Type', mime), ('Content-Length', len(body))])
        return r

def jsonrequest(f):
    """ Decorator marking the decorated method as being a handler for a
    JSON-RPC request (the exact request path is specified via the
    ``$(Controller._cp_path)/$methodname`` combination.

    If the method is called, it will be provided with a :class:`JsonRequest`
    instance and all ``params`` sent during the JSON-RPC request, apart from
    the ``session_id``, ``context`` and ``debug`` keys (which are stripped out
    beforehand)
    """
    f.exposed = 'json'
    return f

class HttpRequest(WebRequest):
    """ Regular GET/POST request
    """
    def dispatch(self, method):
        params = dict(self.httprequest.args)
        params.update(self.httprequest.form)
        params.update(self.httprequest.files)
        self.init(params)
        akw = {}
        for key, value in self.httprequest.args.iteritems():
            if isinstance(value, basestring) and len(value) < 1024:
                akw[key] = value
            else:
                akw[key] = type(value)
        _logger.debug("%s --> %s.%s %r", self.httprequest.method, method.im_class.__name__, method.__name__, akw)
        try:
            r = method(self, **self.params)
        except xmlrpclib.Fault, e:
            r = werkzeug.exceptions.InternalServerError(cgi.escape(simplejson.dumps({
                'code': 200,
                'message': "OpenERP Server Error",
                'data': {
                    'type': 'server_exception',
                    'fault_code': e.faultCode,
                    'debug': "Server %s\nClient %s" % (
                        e.faultString, traceback.format_exc())
                }
            })))
        except Exception:
            logging.getLogger(__name__ + '.HttpRequest.dispatch').exception(
                    "An error occurred while handling a json request")
            r = werkzeug.exceptions.InternalServerError(cgi.escape(simplejson.dumps({
                'code': 300,
                'message': "OpenERP WebClient Error",
                'data': {
                    'type': 'client_exception',
                    'debug': "Client %s" % traceback.format_exc()
                }
            })))
        if self.debug or 1:
            if isinstance(r, (werkzeug.wrappers.BaseResponse, werkzeug.exceptions.HTTPException)):
                _logger.debug('<-- %s', r)
            else:
                _logger.debug("<-- size: %s", len(r))
        return r

    def make_response(self, data, headers=None, cookies=None):
        """ Helper for non-HTML responses, or HTML responses with custom
        response headers or cookies.

        While handlers can just return the HTML markup of a page they want to
        send as a string if non-HTML data is returned they need to create a
        complete response object, or the returned data will not be correctly
        interpreted by the clients.

        :param basestring data: response body
        :param headers: HTTP headers to set on the response
        :type headers: ``[(name, value)]``
        :param collections.Mapping cookies: cookies to set on the client
        """
        response = werkzeug.wrappers.Response(data, headers=headers)
        if cookies:
            for k, v in cookies.iteritems():
                response.set_cookie(k, v)
        return response

    def not_found(self, description=None):
        """ Helper for 404 response, return its result from the method
        """
        return werkzeug.exceptions.NotFound(description)

def httprequest(f):
    """ Decorator marking the decorated method as being a handler for a
    normal HTTP request (the exact request path is specified via the
    ``$(Controller._cp_path)/$methodname`` combination.

    If the method is called, it will be provided with a :class:`HttpRequest`
    instance and all ``params`` sent during the request (``GET`` and ``POST``
    merged in the same dictionary), apart from the ``session_id``, ``context``
    and ``debug`` keys (which are stripped out beforehand)
    """
    f.exposed = 'http'
    return f

#----------------------------------------------------------
# Controller registration with a metaclass
#----------------------------------------------------------
addons_module = {}
addons_manifest = {}
controllers_class = []
controllers_class_path = {}
controllers_object = {}
controllers_object_path = {}
controllers_path = {}

class ControllerType(type):
    def __init__(cls, name, bases, attrs):
        super(ControllerType, cls).__init__(name, bases, attrs)
        name_class = ("%s.%s" % (cls.__module__, cls.__name__), cls)
        controllers_class.append(name_class)
        path = attrs.get('_cp_path')
        if path not in controllers_class_path:
            controllers_class_path[path] = name_class

class Controller(object):
    __metaclass__ = ControllerType

    def __new__(cls, *args, **kwargs):
        subclasses = [c for c in cls.__subclasses__() if c._cp_path == cls._cp_path]
        if subclasses:
            name = "%s (extended by %s)" % (cls.__name__, ', '.join(sub.__name__ for sub in subclasses))
            cls = type(name, tuple(reversed(subclasses)), {})

        return object.__new__(cls)

#----------------------------------------------------------
# Session context manager
#----------------------------------------------------------
@contextlib.contextmanager
def session_context(request, session_store, session_lock, sid):
    with session_lock:
        if sid:
            request.session = session_store.get(sid)
        else:
            request.session = session_store.new()
    try:
        yield request.session
    finally:
        # Remove all OpenERPSession instances with no uid, they're generated
        # either by login process or by HTTP requests without an OpenERP
        # session id, and are generally noise
        removed_sessions = set()
        for key, value in request.session.items():
            if not isinstance(value, session.OpenERPSession):
                continue
            if getattr(value, '_suicide', False) or (
                        not value._uid
                    and not value.jsonp_requests
                    # FIXME do not use a fixed value
                    and value._creation_time + (60*5) < time.time()):
                _logger.debug('remove session %s', key)
                removed_sessions.add(key)
                del request.session[key]

        with session_lock:
            if sid:
                # Re-load sessions from storage and merge non-literal
                # contexts and domains (they're indexed by hash of the
                # content so conflicts should auto-resolve), otherwise if
                # two requests alter those concurrently the last to finish
                # will overwrite the previous one, leading to loss of data
                # (a non-literal is lost even though it was sent to the
                # client and client errors)
                #
                # note that domains_store and contexts_store are append-only (we
                # only ever add items to them), so we can just update one with the
                # other to get the right result, if we want to merge the
                # ``context`` dict we'll need something smarter
                in_store = session_store.get(sid)
                for k, v in request.session.iteritems():
                    stored = in_store.get(k)
                    if stored and isinstance(v, session.OpenERPSession):
                        if hasattr(v, 'contexts_store'):
                            del v.contexts_store
                        if hasattr(v, 'domains_store'):
                            del v.domains_store
                        if not hasattr(v, 'jsonp_requests'):
                            v.jsonp_requests = {}
                        v.jsonp_requests.update(getattr(
                            stored, 'jsonp_requests', {}))

                # add missing keys
                for k, v in in_store.iteritems():
                    if k not in request.session and k not in removed_sessions:
                        request.session[k] = v

            session_store.save(request.session)

def session_gc(session_store):
    if random.random() < 0.001:
        # we keep session one week
        last_week = time.time() - 60*60*24*7
        for fname in os.listdir(session_store.path):
            path = os.path.join(session_store.path, fname)
            try:
                if os.path.getmtime(path) < last_week:
                    os.unlink(path)
            except OSError:
                pass

#----------------------------------------------------------
# WSGI Application
#----------------------------------------------------------
# Add potentially missing (older ubuntu) font mime types
mimetypes.add_type('application/font-woff', '.woff')
mimetypes.add_type('application/vnd.ms-fontobject', '.eot')
mimetypes.add_type('application/x-font-ttf', '.ttf')

class DisableCacheMiddleware(object):
    def __init__(self, app):
        self.app = app
    def __call__(self, environ, start_response):
        def start_wrapped(status, headers):
            referer = environ.get('HTTP_REFERER', '')
            parsed = urlparse.urlparse(referer)
            debug = parsed.query.count('debug') >= 1

            new_headers = []
            unwanted_keys = ['Last-Modified']
            if debug:
                new_headers = [('Cache-Control', 'no-cache')]
                unwanted_keys += ['Expires', 'Etag', 'Cache-Control']

            for k, v in headers:
                if k not in unwanted_keys:
                    new_headers.append((k, v))

            start_response(status, new_headers)
        return self.app(environ, start_wrapped)

def session_path():
    try:
        import pwd
        username = pwd.getpwuid(os.geteuid()).pw_name
    except ImportError:
        try:
            username = getpass.getuser()
        except Exception:
            username = "unknown"
    path = os.path.join(tempfile.gettempdir(), "oe-sessions-" + username)
    try:
        os.mkdir(path, 0700)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            # directory exists: ensure it has the correct permissions
            # this will fail if the directory is not owned by the current user
            os.chmod(path, 0700)
        else:
            raise
    return path

class Root(object):
    """Root WSGI application for the OpenERP Web Client.
    """
    def __init__(self):
        self.addons = {}
        self.statics = {}

        self.load_addons()

        # Setup http sessions
        path = session_path()
        self.session_store = werkzeug.contrib.sessions.FilesystemSessionStore(path)
        self.session_lock = threading.Lock()
        _logger.debug('HTTP sessions stored in: %s', path)

    def __call__(self, environ, start_response):
        """ Handle a WSGI request
        """
        return self.dispatch(environ, start_response)

    def dispatch(self, environ, start_response):
        """
        Performs the actual WSGI dispatching for the application, may be
        wrapped during the initialization of the object.

        Call the object directly.
        """
        request = werkzeug.wrappers.Request(environ)
        request.parameter_storage_class = werkzeug.datastructures.ImmutableDict
        request.app = self

        handler = self.find_handler(*(request.path.split('/')[1:]))

        if not handler:
            response = werkzeug.exceptions.NotFound()
        else:
            sid = request.cookies.get('sid')
            if not sid:
                sid = request.args.get('sid')

            session_gc(self.session_store)

            with session_context(request, self.session_store, self.session_lock, sid) as session:
                result = handler(request)

                if isinstance(result, basestring):
                    headers=[('Content-Type', 'text/html; charset=utf-8'), ('Content-Length', len(result))]
                    response = werkzeug.wrappers.Response(result, headers=headers)
                else:
                    response = result

                if hasattr(response, 'set_cookie'):
                    response.set_cookie('sid', session.sid)

        return response(environ, start_response)

    def load_addons(self):
        """ Load all addons from addons patch containg static files and
        controllers and configure them.  """

        for addons_path in openerp.modules.module.ad_paths:
            for module in sorted(os.listdir(str(addons_path))):
                if module not in addons_module:
                    manifest_path = os.path.join(addons_path, module, '__openerp__.py')
                    path_static = os.path.join(addons_path, module, 'static')
                    if os.path.isfile(manifest_path) and os.path.isdir(path_static):
                        manifest = ast.literal_eval(open(manifest_path).read())
                        manifest['addons_path'] = addons_path
                        _logger.debug("Loading %s", module)
                        if 'openerp.addons' in sys.modules:
                            m = __import__('openerp.addons.' + module)
                        else:
                            m = __import__(module)
                        addons_module[module] = m
                        addons_manifest[module] = manifest
                        self.statics['/%s/static' % module] = path_static

        for k, v in controllers_class_path.items():
            if k not in controllers_object_path and hasattr(v[1], '_cp_path'):
                o = v[1]()
                controllers_object[v[0]] = o
                controllers_object_path[k] = o
                if hasattr(o, '_cp_path'):
                    controllers_path[o._cp_path] = o

        app = werkzeug.wsgi.SharedDataMiddleware(self.dispatch, self.statics)
        self.dispatch = DisableCacheMiddleware(app)

    def find_handler(self, *l):
        """
        Tries to discover the controller handling the request for the path
        specified by the provided parameters

        :param l: path sections to a controller or controller method
        :returns: a callable matching the path sections, or ``None``
        :rtype: ``Controller | None``
        """
        if l:
            ps = '/' + '/'.join(filter(None, l))
            method_name = 'index'
            while ps:
                c = controllers_path.get(ps)
                if c:
                    method = getattr(c, method_name, None)
                    if method:
                        exposed = getattr(method, 'exposed', False)
                        if exposed == 'json':
                            _logger.debug("Dispatch json to %s %s %s", ps, c, method_name)
                            return lambda request: JsonRequest(request).dispatch(method)
                        elif exposed == 'http':
                            _logger.debug("Dispatch http to %s %s %s", ps, c, method_name)
                            return lambda request: HttpRequest(request).dispatch(method)
                ps, _slash, method_name = ps.rpartition('/')
                if not ps and method_name:
                    ps = '/'
        return None

def wsgi_postload():
    openerp.wsgi.register_wsgi_handler(Root())

# vim:et:ts=4:sw=4:



import urlparse
from openerp import sql_db, tools
from qunitsuite.suite import QUnitSuite

class WebSuite(QUnitSuite):
    def __init__(self):
        url = urlparse.urlunsplit([
            'http',
            'localhost:{port}'.format(port=tools.config['xmlrpc_port']),
            '/web/tests',
            'mod=*&source={db}&supadmin={supadmin}&password={password}'.format(
                db=tools.config['db_name'],
                # al: i dont understand why both are needed, db_password is the
                # password for postgres and should not appear here of that i'm
                # sure
                #
                # But runbot provides it with this wrong key so i let it here
                # until it's fixed
                supadmin=tools.config['db_password'] or 'admin',
                password=tools.config['admin_passwd'] or 'admin'),
            ''
        ])
        super(WebSuite, self).__init__(url, 50000)
    def run(self, result):
        if sql_db._Pool is not None:
            sql_db._Pool.close_all(sql_db.dsn(tools.config['db_name']))
        return super(WebSuite, self).run(result)

def load_tests(loader, standard_tests, _):
    standard_tests.addTest(WebSuite())
    return standard_tests



# -*- coding: utf-8 -*-

import ast
import base64
import csv
import glob
import itertools
import logging
import operator
import datetime
import hashlib
import os
import re
import simplejson
import time
import urllib
import urllib2
import urlparse
import xmlrpclib
import zlib
from xml.etree import ElementTree
from cStringIO import StringIO

import babel.messages.pofile
import werkzeug.utils
import werkzeug.wrappers
try:
    import xlwt
except ImportError:
    xlwt = None

import openerp
import openerp.modules.registry
from openerp.tools.translate import _
from openerp.tools import config

from .. import http
openerpweb = http

#----------------------------------------------------------
# OpenERP Web helpers
#----------------------------------------------------------

def rjsmin(script):
    """ Minify js with a clever regex.
    Taken from http://opensource.perlig.de/rjsmin
    Apache License, Version 2.0 """
    def subber(match):
        """ Substitution callback """
        groups = match.groups()
        return (
            groups[0] or
            groups[1] or
            groups[2] or
            groups[3] or
            (groups[4] and '\n') or
            (groups[5] and ' ') or
            (groups[6] and ' ') or
            (groups[7] and ' ') or
            ''
        )

    result = re.sub(
        r'([^\047"/\000-\040]+)|((?:(?:\047[^\047\\\r\n]*(?:\\(?:[^\r\n]|\r?'
        r'\n|\r)[^\047\\\r\n]*)*\047)|(?:"[^"\\\r\n]*(?:\\(?:[^\r\n]|\r?\n|'
        r'\r)[^"\\\r\n]*)*"))[^\047"/\000-\040]*)|(?:(?<=[(,=:\[!&|?{};\r\n]'
        r')(?:[\000-\011\013\014\016-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/'
        r'))*((?:/(?![\r\n/*])[^/\\\[\r\n]*(?:(?:\\[^\r\n]|(?:\[[^\\\]\r\n]*'
        r'(?:\\[^\r\n][^\\\]\r\n]*)*\]))[^/\\\[\r\n]*)*/)[^\047"/\000-\040]*'
        r'))|(?:(?<=[\000-#%-,./:-@\[-^`{-~-]return)(?:[\000-\011\013\014\01'
        r'6-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/))*((?:/(?![\r\n/*])[^/'
        r'\\\[\r\n]*(?:(?:\\[^\r\n]|(?:\[[^\\\]\r\n]*(?:\\[^\r\n][^\\\]\r\n]'
        r'*)*\]))[^/\\\[\r\n]*)*/)[^\047"/\000-\040]*))|(?<=[^\000-!#%&(*,./'
        r':-@\[\\^`{|~])(?:[\000-\011\013\014\016-\040]|(?:/\*[^*]*\*+(?:[^/'
        r'*][^*]*\*+)*/))*(?:((?:(?://[^\r\n]*)?[\r\n]))(?:[\000-\011\013\01'
        r'4\016-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/))*)+(?=[^\000-\040"#'
        r'%-\047)*,./:-@\\-^`|-~])|(?<=[^\000-#%-,./:-@\[-^`{-~-])((?:[\000-'
        r'\011\013\014\016-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)))+(?=[^'
        r'\000-#%-,./:-@\[-^`{-~-])|(?<=\+)((?:[\000-\011\013\014\016-\040]|'
        r'(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)))+(?=\+)|(?<=-)((?:[\000-\011\0'
        r'13\014\016-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/)))+(?=-)|(?:[\0'
        r'00-\011\013\014\016-\040]|(?:/\*[^*]*\*+(?:[^/*][^*]*\*+)*/))+|(?:'
        r'(?:(?://[^\r\n]*)?[\r\n])(?:[\000-\011\013\014\016-\040]|(?:/\*[^*'
        r']*\*+(?:[^/*][^*]*\*+)*/))*)+', subber, '\n%s\n' % script
    ).strip()
    return result

def db_list(req, force=False):
    proxy = req.session.proxy("db")
    dbs = proxy.list(force)
    h = req.httprequest.environ['HTTP_HOST'].split(':')[0]
    d = h.split('.')[0]
    r = openerp.tools.config['dbfilter'].replace('%h', h).replace('%d', d)
    dbs = [i for i in dbs if re.match(r, i)]
    return dbs

def db_monodb_redirect(req):
    return db_redirect(req, not config['list_db'])

def db_redirect(req, match_first_only_if_unique):
    db = False
    redirect = False

    # 1 try the db in the url
    db_url = req.params.get('db')
    if db_url:
        return (db_url, False)

    dbs = db_list(req, True)

    # 2 use the database from the cookie if it's listable and still listed
    cookie_db = req.httprequest.cookies.get('last_used_database')
    if cookie_db in dbs:
        db = cookie_db

    # 3 use the first db if user can list databases
    if dbs and not db and (not match_first_only_if_unique or len(dbs) == 1):
        db = dbs[0]

    # redirect to the chosen db if multiple are available
    if db and len(dbs) > 1:
        query = dict(urlparse.parse_qsl(req.httprequest.query_string, keep_blank_values=True))
        query.update({'db': db})
        redirect = req.httprequest.path + '?' + urllib.urlencode(query)
    return (db, redirect)

def db_monodb(req):
    # if only one db exists, return it else return False
    return db_redirect(req, True)[0]

def redirect_with_hash(req, url, code=303):
    # Most IE and Safari versions decided not to preserve location.hash upon
    # redirect. And even if IE10 pretends to support it, it still fails
    # inexplicably in case of multiple redirects (and we do have some).
    # See extensive test page at http://greenbytes.de/tech/tc/httpredirects/
    return "<html><head><script>window.location = '%s' + location.hash;</script></head></html>" % url

def module_topological_sort(modules):
    """ Return a list of module names sorted so that their dependencies of the
    modules are listed before the module itself

    modules is a dict of {module_name: dependencies}

    :param modules: modules to sort
    :type modules: dict
    :returns: list(str)
    """

    dependencies = set(itertools.chain.from_iterable(modules.itervalues()))
    # incoming edge: dependency on other module (if a depends on b, a has an
    # incoming edge from b, aka there's an edge from b to a)
    # outgoing edge: other module depending on this one

    # [Tarjan 1976], http://en.wikipedia.org/wiki/Topological_sorting#Algorithms
    #L ← Empty list that will contain the sorted nodes
    L = []
    #S ← Set of all nodes with no outgoing edges (modules on which no other
    #    module depends)
    S = set(module for module in modules if module not in dependencies)

    visited = set()
    #function visit(node n)
    def visit(n):
        #if n has not been visited yet then
        if n not in visited:
            #mark n as visited
            visited.add(n)
            #change: n not web module, can not be resolved, ignore
            if n not in modules: return
            #for each node m with an edge from m to n do (dependencies of n)
            for m in modules[n]:
                #visit(m)
                visit(m)
            #add n to L
            L.append(n)
    #for each node n in S do
    for n in S:
        #visit(n)
        visit(n)
    return L

def module_installed(req):
    # Candidates module the current heuristic is the /static dir
    loadable = openerpweb.addons_manifest.keys()
    modules = {}

    # Retrieve database installed modules
    # TODO The following code should move to ir.module.module.list_installed_modules()
    Modules = req.session.model('ir.module.module')
    domain = [('state','=','installed'), ('name','in', loadable)]
    for module in Modules.search_read(domain, ['name', 'dependencies_id']):
        modules[module['name']] = []
        deps = module.get('dependencies_id')
        if deps:
            deps_read = req.session.model('ir.module.module.dependency').read(deps, ['name'])
            dependencies = [i['name'] for i in deps_read]
            modules[module['name']] = dependencies

    sorted_modules = module_topological_sort(modules)
    return sorted_modules

def module_installed_bypass_session(dbname):
    loadable = openerpweb.addons_manifest.keys()
    modules = {}
    try:
        registry = openerp.modules.registry.RegistryManager.get(dbname)
        with registry.cursor() as cr:
            m = registry.get('ir.module.module')
            # TODO The following code should move to ir.module.module.list_installed_modules()
            domain = [('state','=','installed'), ('name','in', loadable)]
            ids = m.search(cr, 1, [('state','=','installed'), ('name','in', loadable)])
            for module in m.read(cr, 1, ids, ['name', 'dependencies_id']):
                modules[module['name']] = []
                deps = module.get('dependencies_id')
                if deps:
                    deps_read = registry.get('ir.module.module.dependency').read(cr, 1, deps, ['name'])
                    dependencies = [i['name'] for i in deps_read]
                    modules[module['name']] = dependencies
    except Exception,e:
        pass
    sorted_modules = module_topological_sort(modules)
    return sorted_modules

def module_boot(req, db=None):
    server_wide_modules = openerp.conf.server_wide_modules or ['web']
    serverside = []
    dbside = []
    for i in server_wide_modules:
        if i in openerpweb.addons_manifest:
            serverside.append(i)
    monodb = db or db_monodb(req)
    if monodb:
        dbside = module_installed_bypass_session(monodb)
        dbside = [i for i in dbside if i not in serverside]
    addons = serverside + dbside
    return addons

def concat_xml(file_list):
    """Concatenate xml files

    :param list(str) file_list: list of files to check
    :returns: (concatenation_result, checksum)
    :rtype: (str, str)
    """
    checksum = hashlib.new('sha1')
    if not file_list:
        return '', checksum.hexdigest()

    root = None
    for fname in file_list:
        with open(fname, 'rb') as fp:
            contents = fp.read()
            checksum.update(contents)
            fp.seek(0)
            xml = ElementTree.parse(fp).getroot()

        if root is None:
            root = ElementTree.Element(xml.tag)
        #elif root.tag != xml.tag:
        #    raise ValueError("Root tags missmatch: %r != %r" % (root.tag, xml.tag))

        for child in xml.getchildren():
            root.append(child)
    return ElementTree.tostring(root, 'utf-8'), checksum.hexdigest()

def concat_files(file_list, reader=None, intersperse=""):
    """ Concatenates contents of all provided files

    :param list(str) file_list: list of files to check
    :param function reader: reading procedure for each file
    :param str intersperse: string to intersperse between file contents
    :returns: (concatenation_result, checksum)
    :rtype: (str, str)
    """
    checksum = hashlib.new('sha1')
    if not file_list:
        return '', checksum.hexdigest()

    if reader is None:
        def reader(f):
            import codecs
            with codecs.open(f, 'rb', "utf-8-sig") as fp:
                return fp.read().encode("utf-8")

    files_content = []
    for fname in file_list:
        contents = reader(fname)
        checksum.update(contents)
        files_content.append(contents)

    files_concat = intersperse.join(files_content)
    return files_concat, checksum.hexdigest()

concat_js_cache = {}

def concat_js(file_list):
    content, checksum = concat_files(file_list, intersperse=';')
    if checksum in concat_js_cache:
        content = concat_js_cache[checksum]
    else:
        content = rjsmin(content)
        concat_js_cache[checksum] = content
    return content, checksum

def fs2web(path):
    """convert FS path into web path"""
    return '/'.join(path.split(os.path.sep))

def manifest_glob(req, extension, addons=None, db=None):
    if addons is None:
        addons = module_boot(req, db=db)
    else:
        addons = addons.split(',')
    r = []
    for addon in addons:
        manifest = openerpweb.addons_manifest.get(addon, None)
        if not manifest:
            continue
        # ensure does not ends with /
        addons_path = os.path.join(manifest['addons_path'], '')[:-1]
        globlist = manifest.get(extension, [])
        for pattern in globlist:
            for path in glob.glob(os.path.normpath(os.path.join(addons_path, addon, pattern))):
                r.append((path, fs2web(path[len(addons_path):])))
    return r

def manifest_list(req, extension, mods=None, db=None):
    """ list ressources to load specifying either:
    mods: a comma separated string listing modules
    db: a database name (return all installed modules in that database)
    """
    if not req.debug:
        path = '/web/webclient/' + extension
        if mods is not None:
            path += '?' + urllib.urlencode({'mods': mods})
        elif db:
            path += '?' + urllib.urlencode({'db': db})
        return [path]
    files = manifest_glob(req, extension, addons=mods, db=db)
    return [wp for _fp, wp in files]

def get_last_modified(files):
    """ Returns the modification time of the most recently modified
    file provided

    :param list(str) files: names of files to check
    :return: most recent modification time amongst the fileset
    :rtype: datetime.datetime
    """
    files = list(files)
    if files:
        return max(datetime.datetime.fromtimestamp(os.path.getmtime(f))
                   for f in files)
    return datetime.datetime(1970, 1, 1)

def make_conditional(req, response, last_modified=None, etag=None):
    """ Makes the provided response conditional based upon the request,
    and mandates revalidation from clients

    Uses Werkzeug's own :meth:`ETagResponseMixin.make_conditional`, after
    setting ``last_modified`` and ``etag`` correctly on the response object

    :param req: OpenERP request
    :type req: web.common.http.WebRequest
    :param response: Werkzeug response
    :type response: werkzeug.wrappers.Response
    :param datetime.datetime last_modified: last modification date of the response content
    :param str etag: some sort of checksum of the content (deep etag)
    :return: the response object provided
    :rtype: werkzeug.wrappers.Response
    """
    response.cache_control.must_revalidate = True
    response.cache_control.max_age = 0
    if last_modified:
        response.last_modified = last_modified
    if etag:
        response.set_etag(etag)
    return response.make_conditional(req.httprequest)

def login_and_redirect(req, db, login, key, redirect_url='/'):
    wsgienv = req.httprequest.environ
    env = dict(
        base_location=req.httprequest.url_root.rstrip('/'),
        HTTP_HOST=wsgienv['HTTP_HOST'],
        REMOTE_ADDR=wsgienv['REMOTE_ADDR'],
    )
    req.session.authenticate(db, login, key, env)
    return set_cookie_and_redirect(req, redirect_url)

def set_cookie_and_redirect(req, redirect_url):
    redirect = werkzeug.utils.redirect(redirect_url, 303)
    redirect.autocorrect_location_header = False
    cookie_val = urllib2.quote(simplejson.dumps(req.session_id))
    redirect.set_cookie('instance0|session_id', cookie_val)
    return redirect

def load_actions_from_ir_values(req, key, key2, models, meta):
    Values = req.session.model('ir.values')
    actions = Values.get(key, key2, models, meta, req.context)

    return [(id, name, clean_action(req, action))
            for id, name, action in actions]

def clean_action(req, action):
    action.setdefault('flags', {})
    action_type = action.setdefault('type', 'ir.actions.act_window_close')
    if action_type == 'ir.actions.act_window':
        return fix_view_modes(action)
    return action

# I think generate_views,fix_view_modes should go into js ActionManager
def generate_views(action):
    """
    While the server generates a sequence called "views" computing dependencies
    between a bunch of stuff for views coming directly from the database
    (the ``ir.actions.act_window model``), it's also possible for e.g. buttons
    to return custom view dictionaries generated on the fly.

    In that case, there is no ``views`` key available on the action.

    Since the web client relies on ``action['views']``, generate it here from
    ``view_mode`` and ``view_id``.

    Currently handles two different cases:

    * no view_id, multiple view_mode
    * single view_id, single view_mode

    :param dict action: action descriptor dictionary to generate a views key for
    """
    view_id = action.get('view_id') or False
    if isinstance(view_id, (list, tuple)):
        view_id = view_id[0]

    # providing at least one view mode is a requirement, not an option
    view_modes = action['view_mode'].split(',')

    if len(view_modes) > 1:
        if view_id:
            raise ValueError('Non-db action dictionaries should provide '
                             'either multiple view modes or a single view '
                             'mode and an optional view id.\n\n Got view '
                             'modes %r and view id %r for action %r' % (
                view_modes, view_id, action))
        action['views'] = [(False, mode) for mode in view_modes]
        return
    action['views'] = [(view_id, view_modes[0])]

def fix_view_modes(action):
    """ For historical reasons, OpenERP has weird dealings in relation to
    view_mode and the view_type attribute (on window actions):

    * one of the view modes is ``tree``, which stands for both list views
      and tree views
    * the choice is made by checking ``view_type``, which is either
      ``form`` for a list view or ``tree`` for an actual tree view

    This methods simply folds the view_type into view_mode by adding a
    new view mode ``list`` which is the result of the ``tree`` view_mode
    in conjunction with the ``form`` view_type.

    TODO: this should go into the doc, some kind of "peculiarities" section

    :param dict action: an action descriptor
    :returns: nothing, the action is modified in place
    """
    if not action.get('views'):
        generate_views(action)

    if action.pop('view_type', 'form') != 'form':
        return action

    if 'view_mode' in action:
        action['view_mode'] = ','.join(
            mode if mode != 'tree' else 'list'
            for mode in action['view_mode'].split(','))
    action['views'] = [
        [id, mode if mode != 'tree' else 'list']
        for id, mode in action['views']
    ]

    return action

def _local_web_translations(trans_file):
    messages = []
    try:
        with open(trans_file) as t_file:
            po = babel.messages.pofile.read_po(t_file)
    except Exception:
        return
    for x in po:
        if x.id and x.string and "openerp-web" in x.auto_comments:
            messages.append({'id': x.id, 'string': x.string})
    return messages

def xml2json_from_elementtree(el, preserve_whitespaces=False):
    """ xml2json-direct
    Simple and straightforward XML-to-JSON converter in Python
    New BSD Licensed
    http://code.google.com/p/xml2json-direct/
    """
    res = {}
    if el.tag[0] == "{":
        ns, name = el.tag.rsplit("}", 1)
        res["tag"] = name
        res["namespace"] = ns[1:]
    else:
        res["tag"] = el.tag
    res["attrs"] = {}
    for k, v in el.items():
        res["attrs"][k] = v
    kids = []
    if el.text and (preserve_whitespaces or el.text.strip() != ''):
        kids.append(el.text)
    for kid in el:
        kids.append(xml2json_from_elementtree(kid, preserve_whitespaces))
        if kid.tail and (preserve_whitespaces or kid.tail.strip() != ''):
            kids.append(kid.tail)
    res["children"] = kids
    return res

def content_disposition(filename, req):
    filename = filename.encode('utf8')
    escaped = urllib2.quote(filename)
    browser = req.httprequest.user_agent.browser
    version = int((req.httprequest.user_agent.version or '0').split('.')[0])
    if browser == 'msie' and version < 9:
        return "attachment; filename=%s" % escaped
    elif browser == 'safari':
        return "attachment; filename=%s" % filename
    else:
        return "attachment; filename*=UTF-8''%s" % escaped


#----------------------------------------------------------
# OpenERP Web web Controllers
#----------------------------------------------------------

html_template = """<!DOCTYPE html>
<html style="height: 100%%">
    <head>
        <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
        <meta http-equiv="content-type" content="text/html; charset=utf-8" />
        <title>OpenERP</title>
        <link rel="shortcut icon" href="/web/static/src/img/favicon.ico" type="image/x-icon"/>
        <link rel="stylesheet" href="/web/static/src/css/full.css" />
        %(css)s
        %(js)s
        <script type="text/javascript">
            $(function() {
                var s = new openerp.init(%(modules)s);
                %(init)s
            });
        </script>
    </head>
    <body>
        <!--[if lte IE 8]>
        <script src="//ajax.googleapis.com/ajax/libs/chrome-frame/1/CFInstall.min.js"></script>
        <script>CFInstall.check({mode: "overlay"});</script>
        <![endif]-->
    </body>
</html>
"""

class Home(openerpweb.Controller):
    _cp_path = '/'

    @openerpweb.httprequest
    def index(self, req, s_action=None, db=None, **kw):
        db, redir = db_monodb_redirect(req)
        if redir:
            return redirect_with_hash(req, redir)

        js = "\n        ".join('<script type="text/javascript" src="%s"></script>' % i for i in manifest_list(req, 'js', db=db))
        css = "\n        ".join('<link rel="stylesheet" href="%s">' % i for i in manifest_list(req, 'css', db=db))

        r = html_template % {
            'js': js,
            'css': css,
            'modules': simplejson.dumps(module_boot(req, db=db)),
            'init': 'var wc = new s.web.WebClient();wc.appendTo($(document.body));'
        }
        return r

    @openerpweb.httprequest
    def login(self, req, db, login, key):
        return login_and_redirect(req, db, login, key)

class WebClient(openerpweb.Controller):
    _cp_path = "/web/webclient"

    @openerpweb.jsonrequest
    def csslist(self, req, mods=None):
        return manifest_list(req, 'css', mods=mods)

    @openerpweb.jsonrequest
    def jslist(self, req, mods=None):
        return manifest_list(req, 'js', mods=mods)

    @openerpweb.jsonrequest
    def qweblist(self, req, mods=None):
        return manifest_list(req, 'qweb', mods=mods)

    @openerpweb.httprequest
    def css(self, req, mods=None, db=None):
        files = list(manifest_glob(req, 'css', addons=mods, db=db))
        last_modified = get_last_modified(f[0] for f in files)
        if req.httprequest.if_modified_since and req.httprequest.if_modified_since >= last_modified:
            return werkzeug.wrappers.Response(status=304)

        file_map = dict(files)

        rx_import = re.compile(r"""@import\s+('|")(?!'|"|/|https?://)""", re.U)
        rx_url = re.compile(r"""url\s*\(\s*('|"|)(?!'|"|/|https?://|data:)""", re.U)

        def reader(f):
            """read the a css file and absolutify all relative uris"""
            with open(f, 'rb') as fp:
                data = fp.read().decode('utf-8')

            path = file_map[f]
            web_dir = os.path.dirname(path)

            data = re.sub(
                rx_import,
                r"""@import \1%s/""" % (web_dir,),
                data,
            )

            data = re.sub(
                rx_url,
                r"""url(\1%s/""" % (web_dir,),
                data,
            )
            return data.encode('utf-8')

        content, checksum = concat_files((f[0] for f in files), reader)

        # move up all @import and @charset rules to the top
        matches = []
        def push(matchobj):
            matches.append(matchobj.group(0))
            return ''

        content = re.sub(re.compile("(@charset.+;$)", re.M), push, content)
        content = re.sub(re.compile("(@import.+;$)", re.M), push, content)

        matches.append(content)
        content = '\n'.join(matches)

        return make_conditional(
            req, req.make_response(content, [('Content-Type', 'text/css')]),
            last_modified, checksum)

    @openerpweb.httprequest
    def js(self, req, mods=None, db=None):
        files = [f[0] for f in manifest_glob(req, 'js', addons=mods, db=db)]
        last_modified = get_last_modified(files)
        if req.httprequest.if_modified_since and req.httprequest.if_modified_since >= last_modified:
            return werkzeug.wrappers.Response(status=304)

        content, checksum = concat_js(files)

        return make_conditional(
            req, req.make_response(content, [('Content-Type', 'application/javascript')]),
            last_modified, checksum)

    @openerpweb.httprequest
    def qweb(self, req, mods=None, db=None):
        files = [f[0] for f in manifest_glob(req, 'qweb', addons=mods, db=db)]
        last_modified = get_last_modified(files)
        if req.httprequest.if_modified_since and req.httprequest.if_modified_since >= last_modified:
            return werkzeug.wrappers.Response(status=304)

        content, checksum = concat_xml(files)

        return make_conditional(
            req, req.make_response(content, [('Content-Type', 'text/xml')]),
            last_modified, checksum)

    @openerpweb.jsonrequest
    def bootstrap_translations(self, req, mods):
        """ Load local translations from *.po files, as a temporary solution
            until we have established a valid session. This is meant only
            for translating the login page and db management chrome, using
            the browser's language. """
        # For performance reasons we only load a single translation, so for
        # sub-languages (that should only be partially translated) we load the
        # main language PO instead - that should be enough for the login screen.
        lang = req.lang.split('_')[0]

        translations_per_module = {}
        for addon_name in mods:
            if openerpweb.addons_manifest[addon_name].get('bootstrap'):
                addons_path = openerpweb.addons_manifest[addon_name]['addons_path']
                f_name = os.path.join(addons_path, addon_name, "i18n", lang + ".po")
                if not os.path.exists(f_name):
                    continue
                translations_per_module[addon_name] = {'messages': _local_web_translations(f_name)}

        return {"modules": translations_per_module,
                "lang_parameters": None}

    @openerpweb.jsonrequest
    def translations(self, req, mods, lang):
        res_lang = req.session.model('res.lang')
        ids = res_lang.search([("code", "=", lang)])
        lang_params = None
        if ids:
            lang_params = res_lang.read(ids[0], ["direction", "date_format", "time_format",
                                                "grouping", "decimal_point", "thousands_sep"])

        # Regional languages (ll_CC) must inherit/override their parent lang (ll), but this is
        # done server-side when the language is loaded, so we only need to load the user's lang.
        ir_translation = req.session.model('ir.translation')
        translations_per_module = {}
        messages = ir_translation.search_read([('module','in',mods),('lang','=',lang),
                                               ('comments','like','openerp-web'),('value','!=',False),
                                               ('value','!=','')],
                                              ['module','src','value','lang'], order='module')
        for mod, msg_group in itertools.groupby(messages, key=operator.itemgetter('module')):
            translations_per_module.setdefault(mod,{'messages':[]})
            translations_per_module[mod]['messages'].extend({'id': m['src'],
                                                             'string': m['value']} \
                                                                for m in msg_group)
        return {"modules": translations_per_module,
                "lang_parameters": lang_params}

    @openerpweb.jsonrequest
    def version_info(self, req):
        return openerp.service.web_services.RPC_VERSION_1

class Proxy(openerpweb.Controller):
    _cp_path = '/web/proxy'

    @openerpweb.jsonrequest
    def load(self, req, path):
        """ Proxies an HTTP request through a JSON request.

        It is strongly recommended to not request binary files through this,
        as the result will be a binary data blob as well.

        :param req: OpenERP request
        :param path: actual request path
        :return: file content
        """
        from werkzeug.test import Client
        from werkzeug.wrappers import BaseResponse

        return Client(req.httprequest.app, BaseResponse).get(path).data

class Database(openerpweb.Controller):
    _cp_path = "/web/database"

    @openerpweb.jsonrequest
    def get_list(self, req):
        # TODO change js to avoid calling this method if in monodb mode
        try:
            return db_list(req)
        except xmlrpclib.Fault:
            monodb = db_monodb(req)
            if monodb:
                return [monodb]
            raise

    @openerpweb.jsonrequest
    def create(self, req, fields):
        params = dict(map(operator.itemgetter('name', 'value'), fields))
        return req.session.proxy("db").create_database(
            params['super_admin_pwd'],
            params['db_name'],
            bool(params.get('demo_data')),
            params['db_lang'],
            params['create_admin_pwd'])

    @openerpweb.jsonrequest
    def duplicate(self, req, fields):
        params = dict(map(operator.itemgetter('name', 'value'), fields))
        duplicate_attrs = (
            params['super_admin_pwd'],
            params['db_original_name'],
            params['db_name'],
        )
        return req.session.proxy("db").duplicate_database(*duplicate_attrs)

    @openerpweb.jsonrequest
    def drop(self, req, fields):
        password, db = operator.itemgetter(
            'drop_pwd', 'drop_db')(
                dict(map(operator.itemgetter('name', 'value'), fields)))
        
        try:
            if req.session.proxy("db").drop(password, db):return True
        except xmlrpclib.Fault, e:
            if e.faultCode and e.faultCode.split(':')[0] == 'AccessDenied':
                return {'error': e.faultCode, 'title': 'Drop Database'}
        return {'error': _('Could not drop database !'), 'title': _('Drop Database')}

    @openerpweb.httprequest
    def backup(self, req, backup_db, backup_pwd, token):
        try:
            db_dump = base64.b64decode(
                req.session.proxy("db").dump(backup_pwd, backup_db))
            filename = "%(db)s_%(timestamp)s.dump" % {
                'db': backup_db,
                'timestamp': datetime.datetime.utcnow().strftime(
                    "%Y-%m-%d_%H-%M-%SZ")
            }
            return req.make_response(db_dump,
               [('Content-Type', 'application/octet-stream; charset=binary'),
               ('Content-Disposition', content_disposition(filename, req))],
               {'fileToken': token}
            )
        except xmlrpclib.Fault, e:
            return simplejson.dumps([[],[{'error': e.faultCode, 'title': _('Backup Database')}]])

    @openerpweb.httprequest
    def restore(self, req, db_file, restore_pwd, new_db):
        try:
            data = base64.b64encode(db_file.read())
            req.session.proxy("db").restore(restore_pwd, new_db, data)
            return ''
        except xmlrpclib.Fault, e:
            if e.faultCode and e.faultCode.split(':')[0] == 'AccessDenied':
                raise Exception("AccessDenied")

    @openerpweb.jsonrequest
    def change_password(self, req, fields):
        old_password, new_password = operator.itemgetter(
            'old_pwd', 'new_pwd')(
                dict(map(operator.itemgetter('name', 'value'), fields)))
        try:
            return req.session.proxy("db").change_admin_password(old_password, new_password)
        except xmlrpclib.Fault, e:
            if e.faultCode and e.faultCode.split(':')[0] == 'AccessDenied':
                return {'error': e.faultCode, 'title': _('Change Password')}
        return {'error': _('Error, password not changed !'), 'title': _('Change Password')}

class Session(openerpweb.Controller):
    _cp_path = "/web/session"

    def session_info(self, req):
        req.session.ensure_valid()
        return {
            "session_id": req.session_id,
            "uid": req.session._uid,
            "user_context": req.session.get_context() if req.session._uid else {},
            "db": req.session._db,
            "username": req.session._login,
        }

    @openerpweb.jsonrequest
    def get_session_info(self, req):
        return self.session_info(req)

    @openerpweb.jsonrequest
    def authenticate(self, req, db, login, password, base_location=None):
        wsgienv = req.httprequest.environ
        env = dict(
            base_location=base_location,
            HTTP_HOST=wsgienv['HTTP_HOST'],
            REMOTE_ADDR=wsgienv['REMOTE_ADDR'],
        )
        req.session.authenticate(db, login, password, env)

        return self.session_info(req)

    @openerpweb.jsonrequest
    def change_password (self,req,fields):
        old_password, new_password,confirm_password = operator.itemgetter('old_pwd', 'new_password','confirm_pwd')(
                dict(map(operator.itemgetter('name', 'value'), fields)))
        if not (old_password.strip() and new_password.strip() and confirm_password.strip()):
            return {'error':_('You cannot leave any password empty.'),'title': _('Change Password')}
        if new_password != confirm_password:
            return {'error': _('The new password and its confirmation must be identical.'),'title': _('Change Password')}
        try:
            if req.session.model('res.users').change_password(
                old_password, new_password):
                return {'new_password':new_password}
        except Exception:
            return {'error': _('The old password you provided is incorrect, your password was not changed.'), 'title': _('Change Password')}
        return {'error': _('Error, password not changed !'), 'title': _('Change Password')}

    @openerpweb.jsonrequest
    def sc_list(self, req):
        return req.session.model('ir.ui.view_sc').get_sc(
            req.session._uid, "ir.ui.menu", req.context)

    @openerpweb.jsonrequest
    def get_lang_list(self, req):
        try:
            return req.session.proxy("db").list_lang() or []
        except Exception, e:
            return {"error": e, "title": _("Languages")}

    @openerpweb.jsonrequest
    def modules(self, req):
        # return all installed modules. Web client is smart enough to not load a module twice
        return module_installed(req)

    @openerpweb.jsonrequest
    def save_session_action(self, req, the_action):
        """
        This method store an action object in the session object and returns an integer
        identifying that action. The method get_session_action() can be used to get
        back the action.

        :param the_action: The action to save in the session.
        :type the_action: anything
        :return: A key identifying the saved action.
        :rtype: integer
        """
        saved_actions = req.httpsession.get('saved_actions')
        if not saved_actions:
            saved_actions = {"next":1, "actions":{}}
            req.httpsession['saved_actions'] = saved_actions
        # we don't allow more than 10 stored actions
        if len(saved_actions["actions"]) >= 10:
            del saved_actions["actions"][min(saved_actions["actions"])]
        key = saved_actions["next"]
        saved_actions["actions"][key] = the_action
        saved_actions["next"] = key + 1
        return key

    @openerpweb.jsonrequest
    def get_session_action(self, req, key):
        """
        Gets back a previously saved action. This method can return None if the action
        was saved since too much time (this case should be handled in a smart way).

        :param key: The key given by save_session_action()
        :type key: integer
        :return: The saved action or None.
        :rtype: anything
        """
        saved_actions = req.httpsession.get('saved_actions')
        if not saved_actions:
            return None
        return saved_actions["actions"].get(key)

    @openerpweb.jsonrequest
    def check(self, req):
        req.session.assert_valid()
        return None

    @openerpweb.jsonrequest
    def destroy(self, req):
        req.session._suicide = True

class Menu(openerpweb.Controller):
    _cp_path = "/web/menu"

    @openerpweb.jsonrequest
    def get_user_roots(self, req):
        """ Return all root menu ids visible for the session user.

        :param req: A request object, with an OpenERP session attribute
        :type req: < session -> OpenERPSession >
        :return: the root menu ids
        :rtype: list(int)
        """
        s = req.session
        Menus = s.model('ir.ui.menu')
        # If a menu action is defined use its domain to get the root menu items
        user_menu_id = s.model('res.users').read([s._uid], ['menu_id'],
                                                 req.context)[0]['menu_id']

        menu_domain = [('parent_id', '=', False)]
        if user_menu_id:
            domain_string = s.model('ir.actions.act_window').read(
                [user_menu_id[0]], ['domain'],req.context)[0]['domain']
            if domain_string:
                menu_domain = ast.literal_eval(domain_string)

        return Menus.search(menu_domain, 0, False, False, req.context)

    @openerpweb.jsonrequest
    def load(self, req):
        """ Loads all menu items (all applications and their sub-menus).

        :param req: A request object, with an OpenERP session attribute
        :type req: < session -> OpenERPSession >
        :return: the menu root
        :rtype: dict('children': menu_nodes)
        """
        Menus = req.session.model('ir.ui.menu')

        fields = ['name', 'sequence', 'parent_id', 'action']
        menu_root_ids = self.get_user_roots(req)
        menu_roots = Menus.read(menu_root_ids, fields, req.context) if menu_root_ids else []
        menu_root = {
            'id': False,
            'name': 'root',
            'parent_id': [-1, ''],
            'children': menu_roots,
            'all_menu_ids': menu_root_ids,
        }
        if not menu_roots:
            return menu_root

        # menus are loaded fully unlike a regular tree view, cause there are a
        # limited number of items (752 when all 6.1 addons are installed)
        menu_ids = Menus.search([('id', 'child_of', menu_root_ids)], 0, False, False, req.context)
        menu_items = Menus.read(menu_ids, fields, req.context)
        # adds roots at the end of the sequence, so that they will overwrite
        # equivalent menu items from full menu read when put into id:item
        # mapping, resulting in children being correctly set on the roots.
        menu_items.extend(menu_roots)
        menu_root['all_menu_ids'] = menu_ids # includes menu_root_ids!

        # make a tree using parent_id
        menu_items_map = dict(
            (menu_item["id"], menu_item) for menu_item in menu_items)
        for menu_item in menu_items:
            if menu_item['parent_id']:
                parent = menu_item['parent_id'][0]
            else:
                parent = False
            if parent in menu_items_map:
                menu_items_map[parent].setdefault(
                    'children', []).append(menu_item)

        # sort by sequence a tree using parent_id
        for menu_item in menu_items:
            menu_item.setdefault('children', []).sort(
                key=operator.itemgetter('sequence'))

        return menu_root

    @openerpweb.jsonrequest
    def load_needaction(self, req, menu_ids):
        """ Loads needaction counters for specific menu ids.

            :return: needaction data
            :rtype: dict(menu_id: {'needaction_enabled': boolean, 'needaction_counter': int})
        """
        return req.session.model('ir.ui.menu').get_needaction_data(menu_ids, req.context)

    @openerpweb.jsonrequest
    def action(self, req, menu_id):
        # still used by web_shortcut
        actions = load_actions_from_ir_values(req,'action', 'tree_but_open',
                                             [('ir.ui.menu', menu_id)], False)
        return {"action": actions}

class DataSet(openerpweb.Controller):
    _cp_path = "/web/dataset"

    @openerpweb.jsonrequest
    def search_read(self, req, model, fields=False, offset=0, limit=False, domain=None, sort=None):
        return self.do_search_read(req, model, fields, offset, limit, domain, sort)
    def do_search_read(self, req, model, fields=False, offset=0, limit=False, domain=None
                       , sort=None):
        """ Performs a search() followed by a read() (if needed) using the
        provided search criteria

        :param req: a JSON-RPC request object
        :type req: openerpweb.JsonRequest
        :param str model: the name of the model to search on
        :param fields: a list of the fields to return in the result records
        :type fields: [str]
        :param int offset: from which index should the results start being returned
        :param int limit: the maximum number of records to return
        :param list domain: the search domain for the query
        :param list sort: sorting directives
        :returns: A structure (dict) with two keys: ids (all the ids matching
                  the (domain, context) pair) and records (paginated records
                  matching fields selection set)
        :rtype: list
        """
        Model = req.session.model(model)

        ids = Model.search(domain, offset or 0, limit or False, sort or False,
                           req.context)
        if limit and len(ids) == limit:
            length = Model.search_count(domain, req.context)
        else:
            length = len(ids) + (offset or 0)
        if fields and fields == ['id']:
            # shortcut read if we only want the ids
            return {
                'length': length,
                'records': [{'id': id} for id in ids]
            }

        records = Model.read(ids, fields or False, req.context)

        index = dict((r['id'], r) for r in records)
        records = [index[x] for x in ids if x in index]

        return {
            'length': length,
            'records': records
        }

    @openerpweb.jsonrequest
    def load(self, req, model, id, fields):
        m = req.session.model(model)
        value = {}
        r = m.read([id], False, req.context)
        if r:
            value = r[0]
        return {'value': value}

    def call_common(self, req, model, method, args, domain_id=None, context_id=None):
        return self._call_kw(req, model, method, args, {})

    def _call_kw(self, req, model, method, args, kwargs):
        # Temporary implements future display_name special field for model#read()
        if method == 'read' and kwargs.get('context', {}).get('future_display_name'):
            if 'display_name' in args[1]:
                names = dict(req.session.model(model).name_get(args[0], **kwargs))
                args[1].remove('display_name')
                records = req.session.model(model).read(*args, **kwargs)
                for record in records:
                    record['display_name'] = \
                        names.get(record['id']) or "%s#%d" % (model, (record['id']))
                return records

        return getattr(req.session.model(model), method)(*args, **kwargs)

    @openerpweb.jsonrequest
    def call(self, req, model, method, args, domain_id=None, context_id=None):
        return self._call_kw(req, model, method, args, {})

    @openerpweb.jsonrequest
    def call_kw(self, req, model, method, args, kwargs):
        return self._call_kw(req, model, method, args, kwargs)

    @openerpweb.jsonrequest
    def call_button(self, req, model, method, args, domain_id=None, context_id=None):
        action = self._call_kw(req, model, method, args, {})
        if isinstance(action, dict) and action.get('type') != '':
            return clean_action(req, action)
        return False

    @openerpweb.jsonrequest
    def exec_workflow(self, req, model, id, signal):
        return req.session.exec_workflow(model, id, signal)

    @openerpweb.jsonrequest
    def resequence(self, req, model, ids, field='sequence', offset=0):
        """ Re-sequences a number of records in the model, by their ids

        The re-sequencing starts at the first model of ``ids``, the sequence
        number is incremented by one after each record and starts at ``offset``

        :param ids: identifiers of the records to resequence, in the new sequence order
        :type ids: list(id)
        :param str field: field used for sequence specification, defaults to
                          "sequence"
        :param int offset: sequence number for first record in ``ids``, allows
                           starting the resequencing from an arbitrary number,
                           defaults to ``0``
        """
        m = req.session.model(model)
        if not m.fields_get([field]):
            return False
        # python 2.6 has no start parameter
        for i, id in enumerate(ids):
            m.write(id, { field: i + offset })
        return True

class View(openerpweb.Controller):
    _cp_path = "/web/view"

    @openerpweb.jsonrequest
    def add_custom(self, req, view_id, arch):
        CustomView = req.session.model('ir.ui.view.custom')
        CustomView.create({
            'user_id': req.session._uid,
            'ref_id': view_id,
            'arch': arch
        }, req.context)
        return {'result': True}

    @openerpweb.jsonrequest
    def undo_custom(self, req, view_id, reset=False):
        CustomView = req.session.model('ir.ui.view.custom')
        vcustom = CustomView.search([('user_id', '=', req.session._uid), ('ref_id' ,'=', view_id)],
                                    0, False, False, req.context)
        if vcustom:
            if reset:
                CustomView.unlink(vcustom, req.context)
            else:
                CustomView.unlink([vcustom[0]], req.context)
            return {'result': True}
        return {'result': False}

class TreeView(View):
    _cp_path = "/web/treeview"

    @openerpweb.jsonrequest
    def action(self, req, model, id):
        return load_actions_from_ir_values(
            req,'action', 'tree_but_open',[(model, id)],
            False)

class Binary(openerpweb.Controller):
    _cp_path = "/web/binary"

    @openerpweb.httprequest
    def image(self, req, model, id, field, **kw):
        last_update = '__last_update'
        Model = req.session.model(model)
        headers = [('Content-Type', 'image/png')]
        etag = req.httprequest.headers.get('If-None-Match')
        hashed_session = hashlib.md5(req.session_id).hexdigest()
        id = None if not id else simplejson.loads(id)
        if type(id) is list:
            id = id[0] # m2o
        if etag:
            if not id and hashed_session == etag:
                return werkzeug.wrappers.Response(status=304)
            else:
                date = Model.read([id], [last_update], req.context)[0].get(last_update)
                if hashlib.md5(date).hexdigest() == etag:
                    return werkzeug.wrappers.Response(status=304)

        retag = hashed_session
        try:
            if not id:
                res = Model.default_get([field], req.context).get(field)
                image_base64 = res
            else:
                res = Model.read([id], [last_update, field], req.context)[0]
                retag = hashlib.md5(res.get(last_update)).hexdigest()
                image_base64 = res.get(field)

            if kw.get('resize'):
                resize = kw.get('resize').split(',')
                if len(resize) == 2 and int(resize[0]) and int(resize[1]):
                    width = int(resize[0])
                    height = int(resize[1])
                    # resize maximum 500*500
                    if width > 500: width = 500
                    if height > 500: height = 500
                    image_base64 = openerp.tools.image_resize_image(base64_source=image_base64, size=(width, height), encoding='base64', filetype='PNG')

            image_data = base64.b64decode(image_base64)

        except (TypeError, xmlrpclib.Fault):
            image_data = self.placeholder(req)
        headers.append(('ETag', retag))
        headers.append(('Content-Length', len(image_data)))
        try:
            ncache = int(kw.get('cache'))
            headers.append(('Cache-Control', 'no-cache' if ncache == 0 else 'max-age=%s' % (ncache)))
        except:
            pass
        return req.make_response(image_data, headers)

    def placeholder(self, req, image='placeholder.png'):
        addons_path = openerpweb.addons_manifest['web']['addons_path']
        return open(os.path.join(addons_path, 'web', 'static', 'src', 'img', image), 'rb').read()

    @openerpweb.httprequest
    def saveas(self, req, model, field, id=None, filename_field=None, **kw):
        """ Download link for files stored as binary fields.

        If the ``id`` parameter is omitted, fetches the default value for the
        binary field (via ``default_get``), otherwise fetches the field for
        that precise record.

        :param req: OpenERP request
        :type req: :class:`web.common.http.HttpRequest`
        :param str model: name of the model to fetch the binary from
        :param str field: binary field
        :param str id: id of the record from which to fetch the binary
        :param str filename_field: field holding the file's name, if any
        :returns: :class:`werkzeug.wrappers.Response`
        """
        Model = req.session.model(model)
        fields = [field]
        if filename_field:
            fields.append(filename_field)
        if id:
            res = Model.read([int(id)], fields, req.context)[0]
        else:
            res = Model.default_get(fields, req.context)
        filecontent = base64.b64decode(res.get(field, ''))
        if not filecontent:
            return req.not_found()
        else:
            filename = '%s_%s' % (model.replace('.', '_'), id)
            if filename_field:
                filename = res.get(filename_field, '') or filename
            return req.make_response(filecontent,
                [('Content-Type', 'application/octet-stream'),
                 ('Content-Disposition', content_disposition(filename, req))])

    @openerpweb.httprequest
    def saveas_ajax(self, req, data, token):
        jdata = simplejson.loads(data)
        model = jdata['model']
        field = jdata['field']
        data = jdata['data']
        id = jdata.get('id', None)
        filename_field = jdata.get('filename_field', None)
        context = jdata.get('context', {})

        Model = req.session.model(model)
        fields = [field]
        if filename_field:
            fields.append(filename_field)
        if data:
            res = { field: data }
        elif id:
            res = Model.read([int(id)], fields, context)[0]
        else:
            res = Model.default_get(fields, context)
        filecontent = base64.b64decode(res.get(field, ''))
        if not filecontent:
            raise ValueError(_("No content found for field '%s' on '%s:%s'") %
                (field, model, id))
        else:
            filename = '%s_%s' % (model.replace('.', '_'), id)
            if filename_field:
                filename = res.get(filename_field, '') or filename
            return req.make_response(filecontent,
                headers=[('Content-Type', 'application/octet-stream'),
                        ('Content-Disposition', content_disposition(filename, req))],
                cookies={'fileToken': token})

    @openerpweb.httprequest
    def upload(self, req, callback, ufile):
        # TODO: might be useful to have a configuration flag for max-length file uploads
        out = """<script language="javascript" type="text/javascript">
                    var win = window.top.window;
                    win.jQuery(win).trigger(%s, %s);
                </script>"""
        try:
            data = ufile.read()
            args = [len(data), ufile.filename,
                    ufile.content_type, base64.b64encode(data)]
        except Exception, e:
            args = [False, e.message]
        return out % (simplejson.dumps(callback), simplejson.dumps(args))

    @openerpweb.httprequest
    def upload_attachment(self, req, callback, model, id, ufile):
        Model = req.session.model('ir.attachment')
        out = """<script language="javascript" type="text/javascript">
                    var win = window.top.window;
                    win.jQuery(win).trigger(%s, %s);
                </script>"""
        try:
            attachment_id = Model.create({
                'name': ufile.filename,
                'datas': base64.encodestring(ufile.read()),
                'datas_fname': ufile.filename,
                'res_model': model,
                'res_id': int(id)
            }, req.context)
            args = {
                'filename': ufile.filename,
                'id':  attachment_id
            }
        except xmlrpclib.Fault, e:
            args = {'error':e.faultCode }
        return out % (simplejson.dumps(callback), simplejson.dumps(args))

    @openerpweb.httprequest
    def company_logo(self, req, dbname=None):
        # TODO add etag, refactor to use /image code for etag
        uid = None
        if req.session._db:
            dbname = req.session._db
            uid = req.session._uid
        elif dbname is None:
            dbname = db_monodb(req)

        if not uid:
            uid = openerp.SUPERUSER_ID

        if not dbname:
            image_data = self.placeholder(req, 'logo.png')
        else:
            try:
                # create an empty registry
                registry = openerp.modules.registry.Registry(dbname)
                with registry.cursor() as cr:
                    cr.execute("""SELECT c.logo_web
                                    FROM res_users u
                               LEFT JOIN res_company c
                                      ON c.id = u.company_id
                                   WHERE u.id = %s
                               """, (uid,))
                    row = cr.fetchone()
                    if row and row[0]:
                        image_data = str(row[0]).decode('base64')
                    else:
                        image_data = self.placeholder(req, 'nologo.png')
            except Exception:
                image_data = self.placeholder(req, 'logo.png')

        headers = [
            ('Content-Type', 'image/png'),
            ('Content-Length', len(image_data)),
        ]
        return req.make_response(image_data, headers)

class Action(openerpweb.Controller):
    _cp_path = "/web/action"

    @openerpweb.jsonrequest
    def load(self, req, action_id, do_not_eval=False):
        Actions = req.session.model('ir.actions.actions')
        value = False
        try:
            action_id = int(action_id)
        except ValueError:
            try:
                module, xmlid = action_id.split('.', 1)
                model, action_id = req.session.model('ir.model.data').get_object_reference(module, xmlid)
                assert model.startswith('ir.actions.')
            except Exception:
                action_id = 0   # force failed read

        base_action = Actions.read([action_id], ['type'], req.context)
        if base_action:
            ctx = {}
            action_type = base_action[0]['type']
            if action_type == 'ir.actions.report.xml':
                ctx.update({'bin_size': True})
            ctx.update(req.context)
            action = req.session.model(action_type).read([action_id], False, ctx)
            if action:
                value = clean_action(req, action[0])
        return value

    @openerpweb.jsonrequest
    def run(self, req, action_id):
        return_action = req.session.model('ir.actions.server').run(
            [action_id], req.context)
        if return_action:
            return clean_action(req, return_action)
        else:
            return False

class Export(openerpweb.Controller):
    _cp_path = "/web/export"

    @openerpweb.jsonrequest
    def formats(self, req):
        """ Returns all valid export formats

        :returns: for each export format, a pair of identifier and printable name
        :rtype: [(str, str)]
        """
        return sorted([
            controller.fmt
            for path, controller in openerpweb.controllers_path.iteritems()
            if path.startswith(self._cp_path)
            if hasattr(controller, 'fmt')
        ], key=operator.itemgetter("label"))

    def fields_get(self, req, model):
        Model = req.session.model(model)
        fields = Model.fields_get(False, req.context)
        return fields

    @openerpweb.jsonrequest
    def get_fields(self, req, model, prefix='', parent_name= '',
                   import_compat=True, parent_field_type=None,
                   exclude=None):

        if import_compat and parent_field_type == "many2one":
            fields = {}
        else:
            fields = self.fields_get(req, model)

        if import_compat:
            fields.pop('id', None)
        else:
            fields['.id'] = fields.pop('id', {'string': 'ID'})

        fields_sequence = sorted(fields.iteritems(),
            key=lambda field: field[1].get('string', ''))

        records = []
        for field_name, field in fields_sequence:
            if import_compat:
                if exclude and field_name in exclude:
                    continue
                if field.get('readonly'):
                    # If none of the field's states unsets readonly, skip the field
                    if all(dict(attrs).get('readonly', True)
                           for attrs in field.get('states', {}).values()):
                        continue
            if not field.get('exportable', True):
                continue

            id = prefix + (prefix and '/'or '') + field_name
            name = parent_name + (parent_name and '/' or '') + field['string']
            record = {'id': id, 'string': name,
                      'value': id, 'children': False,
                      'field_type': field.get('type'),
                      'required': field.get('required'),
                      'relation_field': field.get('relation_field')}
            records.append(record)

            if len(name.split('/')) < 3 and 'relation' in field:
                ref = field.pop('relation')
                record['value'] += '/id'
                record['params'] = {'model': ref, 'prefix': id, 'name': name}

                if not import_compat or field['type'] == 'one2many':
                    # m2m field in import_compat is childless
                    record['children'] = True

        return records

    @openerpweb.jsonrequest
    def namelist(self,req,  model, export_id):
        # TODO: namelist really has no reason to be in Python (although itertools.groupby helps)
        export = req.session.model("ir.exports").read([export_id])[0]
        export_fields_list = req.session.model("ir.exports.line").read(
            export['export_fields'])

        fields_data = self.fields_info(
            req, model, map(operator.itemgetter('name'), export_fields_list))

        return [
            {'name': field_name, 'label': fields_data[field_name]}
            for field_name in fields_data.keys()
        ]

    def fields_info(self, req, model, export_fields):
        info = {}
        fields = self.fields_get(req, model)
        if ".id" in export_fields:
            fields['.id'] = fields.pop('id', {'string': 'ID'})

        # To make fields retrieval more efficient, fetch all sub-fields of a
        # given field at the same time. Because the order in the export list is
        # arbitrary, this requires ordering all sub-fields of a given field
        # together so they can be fetched at the same time
        #
        # Works the following way:
        # * sort the list of fields to export, the default sorting order will
        #   put the field itself (if present, for xmlid) and all of its
        #   sub-fields right after it
        # * then, group on: the first field of the path (which is the same for
        #   a field and for its subfields and the length of splitting on the
        #   first '/', which basically means grouping the field on one side and
        #   all of the subfields on the other. This way, we have the field (for
        #   the xmlid) with length 1, and all of the subfields with the same
        #   base but a length "flag" of 2
        # * if we have a normal field (length 1), just add it to the info
        #   mapping (with its string) as-is
        # * otherwise, recursively call fields_info via graft_subfields.
        #   all graft_subfields does is take the result of fields_info (on the
        #   field's model) and prepend the current base (current field), which
        #   rebuilds the whole sub-tree for the field
        #
        # result: because we're not fetching the fields_get for half the
        # database models, fetching a namelist with a dozen fields (including
        # relational data) falls from ~6s to ~300ms (on the leads model).
        # export lists with no sub-fields (e.g. import_compatible lists with
        # no o2m) are even more efficient (from the same 6s to ~170ms, as
        # there's a single fields_get to execute)
        for (base, length), subfields in itertools.groupby(
                sorted(export_fields),
                lambda field: (field.split('/', 1)[0], len(field.split('/', 1)))):
            subfields = list(subfields)
            if length == 2:
                # subfields is a seq of $base/*rest, and not loaded yet
                info.update(self.graft_subfields(
                    req, fields[base]['relation'], base, fields[base]['string'],
                    subfields
                ))
            elif base in fields:
                info[base] = fields[base]['string']

        return info

    def graft_subfields(self, req, model, prefix, prefix_string, fields):
        export_fields = [field.split('/', 1)[1] for field in fields]
        return (
            (prefix + '/' + k, prefix_string + '/' + v)
            for k, v in self.fields_info(req, model, export_fields).iteritems())

class ExportFormat(object):
    @property
    def content_type(self):
        """ Provides the format's content type """
        raise NotImplementedError()

    def filename(self, base):
        """ Creates a valid filename for the format (with extension) from the
         provided base name (exension-less)
        """
        raise NotImplementedError()

    def from_data(self, fields, rows):
        """ Conversion method from OpenERP's export data to whatever the
        current export class outputs

        :params list fields: a list of fields to export
        :params list rows: a list of records to export
        :returns:
        :rtype: bytes
        """
        raise NotImplementedError()

    @openerpweb.httprequest
    def index(self, req, data, token):
        model, fields, ids, domain, import_compat = \
            operator.itemgetter('model', 'fields', 'ids', 'domain',
                                'import_compat')(
                simplejson.loads(data))

        Model = req.session.model(model)
        ids = ids or Model.search(domain, 0, False, False, req.context)

        field_names = map(operator.itemgetter('name'), fields)
        import_data = Model.export_data(ids, field_names, req.context).get('datas',[])

        if import_compat:
            columns_headers = field_names
        else:
            columns_headers = [val['label'].strip() for val in fields]


        return req.make_response(self.from_data(columns_headers, import_data),
            headers=[('Content-Disposition',
                            content_disposition(self.filename(model), req)),
                     ('Content-Type', self.content_type)],
            cookies={'fileToken': token})

class CSVExport(ExportFormat, http.Controller):
    _cp_path = '/web/export/csv'
    fmt = {'tag': 'csv', 'label': 'CSV'}

    @property
    def content_type(self):
        return 'text/csv;charset=utf8'

    def filename(self, base):
        return base + '.csv'

    def from_data(self, fields, rows):
        fp = StringIO()
        writer = csv.writer(fp, quoting=csv.QUOTE_ALL)

        writer.writerow([name.encode('utf-8') for name in fields])

        for data in rows:
            row = []
            for d in data:
                if isinstance(d, basestring):
                    d = d.replace('\n',' ').replace('\t',' ')
                    try:
                        d = d.encode('utf-8')
                    except UnicodeError:
                        pass
                if d is False: d = None
                row.append(d)
            writer.writerow(row)

        fp.seek(0)
        data = fp.read()
        fp.close()
        return data

class ExcelExport(ExportFormat, http.Controller):
    _cp_path = '/web/export/xls'
    fmt = {
        'tag': 'xls',
        'label': 'Excel',
        'error': None if xlwt else "XLWT required"
    }

    @property
    def content_type(self):
        return 'application/vnd.ms-excel'

    def filename(self, base):
        return base + '.xls'

    def from_data(self, fields, rows):
        workbook = xlwt.Workbook()
        worksheet = workbook.add_sheet('Sheet 1')

        for i, fieldname in enumerate(fields):
            worksheet.write(0, i, fieldname)
            worksheet.col(i).width = 8000 # around 220 pixels

        style = xlwt.easyxf('align: wrap yes')

        for row_index, row in enumerate(rows):
            for cell_index, cell_value in enumerate(row):
                if isinstance(cell_value, basestring):
                    cell_value = re.sub("\r", " ", cell_value)
                if cell_value is False: cell_value = None
                worksheet.write(row_index + 1, cell_index, cell_value, style)

        fp = StringIO()
        workbook.save(fp)
        fp.seek(0)
        data = fp.read()
        fp.close()
        return data

class Reports(openerpweb.Controller):
    _cp_path = "/web/report"
    POLLING_DELAY = 0.25
    TYPES_MAPPING = {
        'doc': 'application/vnd.ms-word',
        'html': 'text/html',
        'odt': 'application/vnd.oasis.opendocument.text',
        'pdf': 'application/pdf',
        'sxw': 'application/vnd.sun.xml.writer',
        'xls': 'application/vnd.ms-excel',
    }

    @openerpweb.httprequest
    def index(self, req, action, token):
        action = simplejson.loads(action)

        report_srv = req.session.proxy("report")
        context = dict(req.context)
        context.update(action["context"])

        report_data = {}
        report_ids = context["active_ids"]
        if 'report_type' in action:
            report_data['report_type'] = action['report_type']
        if 'datas' in action:
            if 'ids' in action['datas']:
                report_ids = action['datas'].pop('ids')
            report_data.update(action['datas'])

        report_id = report_srv.report(
            req.session._db, req.session._uid, req.session._password,
            action["report_name"], report_ids,
            report_data, context)

        report_struct = None
        while True:
            report_struct = report_srv.report_get(
                req.session._db, req.session._uid, req.session._password, report_id)
            if report_struct["state"]:
                break

            time.sleep(self.POLLING_DELAY)

        report = base64.b64decode(report_struct['result'])
        if report_struct.get('code') == 'zlib':
            report = zlib.decompress(report)
        report_mimetype = self.TYPES_MAPPING.get(
            report_struct['format'], 'octet-stream')
        file_name = action.get('name', 'report')
        if 'name' not in action:
            reports = req.session.model('ir.actions.report.xml')
            res_id = reports.search([('report_name', '=', action['report_name']),],
                                    0, False, False, context)
            if len(res_id) > 0:
                file_name = reports.read(res_id[0], ['name'], context)['name']
            else:
                file_name = action['report_name']
        file_name = '%s.%s' % (file_name, report_struct['format'])

        return req.make_response(report,
             headers=[
                 ('Content-Disposition', content_disposition(file_name, req)),
                 ('Content-Type', report_mimetype),
                 ('Content-Length', len(report))],
             cookies={'fileToken': token})

# vim:expandtab:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: utf-8 -*-
import collections
import mock
import unittest2

from ..controllers import main

class Placeholder(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.iteritems():
            setattr(self, k, v)

class LoadTest(unittest2.TestCase):
    def setUp(self):
        self.menu = main.Menu()
        self.request = mock.Mock()

        # Have self.request.session.model() return a different mock object for
        # each model (but always the same mock for a given model name)
        models = collections.defaultdict(mock.Mock)
        model = self.request.session.model.side_effect = \
            lambda model_name: models[model_name]

        self.MockMenus = model('ir.ui.menu')
        # Mock the absence of custom menu
        model('res.users').read.return_value = [{
            'menu_id': False
        }]

    def tearDown(self):
        del self.request
        del self.MockMenus
        del self.menu

    def test_empty(self):
        self.MockMenus.search.return_value = []
        self.MockMenus.read.return_value = []

        root = self.menu.load(self.request)

        self.MockMenus.search.assert_called_with(
            [('parent_id','=', False)], 0, False, False,
            self.request.context)

        self.assertEqual(root['all_menu_ids'], [])

        self.assertListEqual(
            root['children'],
            [])

    def test_applications_sort(self):
        self.MockMenus.search.return_value = [1, 2, 3]
        self.MockMenus.read.side_effect = lambda *args: [
            {'id': 1, 'sequence': 1, 'parent_id': False},
            {'id': 3, 'sequence': 2, 'parent_id': False},
            {'id': 2, 'sequence': 3, 'parent_id': False},
        ]

        root = self.menu.load(self.request)

        self.MockMenus.search.assert_called_with(
            [('id','child_of', [1, 2, 3])], 0, False, False,
            self.request.context)

        self.MockMenus.read.assert_called_with(
            [1, 2, 3], ['name', 'sequence', 'parent_id',
                        'action'],
            self.request.context)

        self.assertEqual(root['all_menu_ids'], [1, 2, 3])

        self.assertEqual(
            root['children'],
            [{
                'id': 1, 'sequence': 1,
                'parent_id': False, 'children': []
            }, {
                'id': 3, 'sequence': 2,
                'parent_id': False, 'children': []
            }, {
                'id': 2, 'sequence': 3,
                'parent_id': False, 'children': []
            }])

    def test_deep(self):
        self.MockMenus.search.side_effect = lambda domain, *args: (
            [1] if domain == [('parent_id', '=', False)] else [1, 2, 3, 4])

        root = {'id': 1, 'sequence': 1, 'parent_id': False}
        self.MockMenus.read.side_effect = lambda ids, *args: (
            [root] if ids == [1] else [
                {'id': 1, 'sequence': 1, 'parent_id': False},
                {'id': 2, 'sequence': 2, 'parent_id': [1, '']},
                {'id': 3, 'sequence': 1, 'parent_id': [2, '']},
                {'id': 4, 'sequence': 2, 'parent_id': [2, '']},
            ])

        root = self.menu.load(self.request)

        self.MockMenus.search.assert_called_with(
            [('id','child_of', [1])], 0, False, False,
            self.request.context)

        self.assertEqual(root['all_menu_ids'], [1, 2, 3, 4])

        self.assertEqual(
            root['children'],
            [{
                 'id': 1,
                 'sequence': 1,
                 'parent_id': False,
                 'children': [{
                     'id': 2,
                     'sequence': 2,
                     'parent_id': [1, ''],
                     'children': [{
                         'id': 3,
                         'sequence': 1,
                         'parent_id': [2, ''],
                         'children': []
                     }, {
                         'id': 4,
                         'sequence': 2,
                         'parent_id': [2, ''],
                         'children': []
                     }]
                 }]
            }]
        )

class ActionMungerTest(unittest2.TestCase):
    def setUp(self):
        self.menu = main.Menu()
    def test_actual_treeview(self):
        action = {
            "views": [[False, "tree"], [False, "form"],
                      [False, "calendar"]],
            "view_type": "tree",
            "view_id": False,
            "view_mode": "tree,form,calendar"
        }
        changed = action.copy()
        del action['view_type']
        main.fix_view_modes(changed)

        self.assertEqual(changed, action)

    def test_list_view(self):
        action = {
            "views": [[False, "tree"], [False, "form"],
                      [False, "calendar"]],
            "view_type": "form",
            "view_id": False,
            "view_mode": "tree,form,calendar"
        }
        main.fix_view_modes(action)

        self.assertEqual(action, {
            "views": [[False, "list"], [False, "form"],
                      [False, "calendar"]],
            "view_id": False,
            "view_mode": "list,form,calendar"
        })

    def test_redundant_views(self):

        action = {
            "views": [[False, "tree"], [False, "form"],
                      [False, "calendar"], [42, "tree"]],
            "view_type": "form",
            "view_id": False,
            "view_mode": "tree,form,calendar"
        }
        main.fix_view_modes(action)

        self.assertEqual(action, {
            "views": [[False, "list"], [False, "form"],
                      [False, "calendar"], [42, "list"]],
            "view_id": False,
            "view_mode": "list,form,calendar"
        })



{
    'name': 'Web',
    'category': 'Hidden',
    'version': '7.0.1.0',
    'description':
        """
OpenERP Web core module.
========================

This module provides the core of the OpenERP Web Client.
        """,
    'depends': [],
    'auto_install': True,
    'post_load': 'wsgi_postload',
    'js' : [
        "static/src/fixbind.js",
        "static/lib/datejs/globalization/en-US.js",
        "static/lib/datejs/core.js",
        "static/lib/datejs/parser.js",
        "static/lib/datejs/sugarpak.js",
        "static/lib/datejs/extras.js",
        "static/lib/jquery/jquery-1.8.3.js",
        "static/lib/jquery.MD5/jquery.md5.js",
        "static/lib/jquery.form/jquery.form.js",
        "static/lib/jquery.validate/jquery.validate.js",
        "static/lib/jquery.ba-bbq/jquery.ba-bbq.js",
        "static/lib/spinjs/spin.js",
        "static/lib/jquery.autosize/jquery.autosize.js",
        "static/lib/jquery.blockUI/jquery.blockUI.js",
        "static/lib/jquery.placeholder/jquery.placeholder.js",
        "static/lib/jquery.ui/js/jquery-ui-1.9.1.custom.js",
        "static/lib/jquery.ui.timepicker/js/jquery-ui-timepicker-addon.js",
        "static/lib/jquery.ui.notify/js/jquery.notify.js",
        "static/lib/jquery.deferred-queue/jquery.deferred-queue.js",
        "static/lib/jquery.scrollTo/jquery.scrollTo-min.js",
        "static/lib/jquery.tipsy/jquery.tipsy.js",
        "static/lib/jquery.textext/jquery.textext.js",
        "static/lib/jquery.timeago/jquery.timeago.js",
        "static/lib/qweb/qweb2.js",
        "static/lib/underscore/underscore.js",
        "static/lib/underscore/underscore.string.js",
        "static/lib/backbone/backbone.js",
        "static/lib/cleditor/jquery.cleditor.js",
        "static/lib/py.js/lib/py.js",
        "static/src/js/boot.js",
        "static/src/js/testing.js",
        "static/src/js/pyeval.js",
        "static/src/js/corelib.js",
        "static/src/js/coresetup.js",
        "static/src/js/dates.js",
        "static/src/js/formats.js",
        "static/src/js/chrome.js",
        "static/src/js/views.js",
        "static/src/js/data.js",
        "static/src/js/data_export.js",
        "static/src/js/search.js",
        "static/src/js/view_form.js",
        "static/src/js/view_list.js",
        "static/src/js/view_list_editable.js",
        "static/src/js/view_tree.js",
    ],
    'css' : [
        "static/lib/jquery.ui.bootstrap/css/custom-theme/jquery-ui-1.9.0.custom.css",
        "static/lib/jquery.ui.timepicker/css/jquery-ui-timepicker-addon.css",
        "static/lib/jquery.ui.notify/css/ui.notify.css",
        "static/lib/jquery.tipsy/tipsy.css",
        "static/lib/jquery.textext/jquery.textext.css",
        "static/src/css/base.css",
        "static/src/css/data_export.css",
        "static/lib/cleditor/jquery.cleditor.css",
    ],
    'qweb' : [
        "static/src/xml/*.xml",
    ],
    'test': [
        "static/test/testing.js",
        "static/test/class.js",
        "static/test/registry.js",
        "static/test/form.js",
        "static/test/data.js",
        "static/test/list-utils.js",
        "static/test/formats.js",
        "static/test/rpc.js",
        "static/test/evals.js",
        "static/test/search.js",
        "static/test/Widget.js",
        "static/test/list.js",
        "static/test/list-editable.js",
        "static/test/mutex.js"
    ],
    'bootstrap': True,
}



import json
import subprocess
import unittest
import os
import time

ROOT = os.path.join(os.path.dirname(__file__), 'grunt')

__all__ = ['QUnitSuite']

def _exc_info_to_string(err, test):
    return err

class QUnitTest(unittest.TestCase):
    def __init__(self, module, name):
        self.module = module
        self.name = name
        self.failed = False
    def shortDescription(self):
        return None
    def __repr__(self):
        return '<QUnitTest %s:%s>' % (self.module, self.name)
    def __str__(self):
        return '%s: %s' % (self.module, self.name)

class QUnitSuite(unittest.TestSuite):
    def __init__(self, qunitfile, timeout=5000):
        self.testfile = qunitfile
        self.timeout = timeout
        self._module = None
        self._test = None

    def __iter__(self):
        return iter([self])

    def run(self, result):
        try:
            subprocess.call(['phantomjs', '-v'],
                            stdout=open(os.devnull, 'w'),
                            stderr=subprocess.STDOUT)
        except OSError:
            test = QUnitTest('phantomjs', 'javascript tests')
            result.startTest(test)
            result.startTest(test)
            result.addSkip(test , "phantomjs command not found")
            result.stopTest(test)
            return

        result._exc_info_to_string = _exc_info_to_string
        try:
            self._run(result)
        finally:
            del result._exc_info_to_string

    def _run(self, result):
        phantom = subprocess.Popen([
            'phantomjs',
            '--config=%s' % os.path.join(ROOT, 'phantomjs.json'),
            os.path.join(ROOT, 'bootstrap.js'), self.testfile,
            json.dumps({
                'timeout': self.timeout,
                'inject': os.path.join(ROOT, 'qunit-phantomjs-bridge.js')
            })
        ], stdout=subprocess.PIPE)

        try:
            while True:
                line = phantom.stdout.readline()
                if line:
                    if self.process(line, result):
                        break
                else:
                    time.sleep(0.1)
        finally:
            # If the phantomjs process hasn't quit, kill it
            if phantom.poll() is None:
                phantom.terminate()

    def process(self, line, result):
        args = json.loads(line)
        event_name = args[0]

        if event_name == 'qunit.done':
            return True
        elif event_name == 'fail.load':
            self.add_error(result, "PhantomJS unable to load %s" % args[1])
            return True
        elif event_name == 'fail.timeout':
            self.add_error(result, "PhantomJS timed out, possibly due to a"
                                   " missing QUnit start() call")
            return True

        elif event_name == 'qunit.moduleStart':
            self._module = args[1].encode('utf-8')
        elif event_name == 'qunit.moduleStop':
            self._test = None
            self._module = None
        elif event_name == 'qunit.testStart':
            self._test = QUnitTest(self._module, args[1].encode('utf-8'))
            result.startTest(self._test)
        elif event_name == 'qunit.testDone':
            if not self._test.failed:
                result.addSuccess(self._test)
            result.stopTest(self._test)
            self._test = None
        elif event_name == 'qunit.log':
            if args[1]:
                return False

            self._test.failed = True
            result.addFailure(
                self._test, self.failure_to_str(*args[2:]))

        return False

    def add_error(self, result, s):
        test = QUnitTest('phantomjs', 'startup')
        result.startTest(test)
        result.addError(test, s)
        result.stopTest(test)

    def failure_to_str(self, actual, expected, message, source):
        if message or actual == expected:
            formatted = str(message or '')
        else:
            formatted = "%s != %s" % (actual, expected)

        if source:
            formatted += '\n\n' + source

        return formatted



# -*- coding: utf-8 -*-

import random
import unittest2

from ..controllers.main import module_topological_sort as sort

def sample(population):
    return random.sample(
        population,
            random.randint(0, min(len(population), 5)))

class TestModulesLoading(unittest2.TestCase):
    def setUp(self):
        self.mods = map(str, range(1000))
    def test_topological_sort(self):
        random.shuffle(self.mods)
        modules = [
            (k, sample(self.mods[:i]))
            for i, k in enumerate(self.mods)]
        random.shuffle(modules)
        ms = dict(modules)

        seen = set()
        sorted_modules = sort(ms)
        for module in sorted_modules:
            deps = ms[module]
            self.assertGreaterEqual(
                seen, set(deps),
                        'Module %s (index %d), ' \
                        'missing dependencies %s from loaded modules %s' % (
                    module, sorted_modules.index(module), deps, seen
                ))
            seen.add(module)



import datetime
import babel
import dateutil.relativedelta
import logging
import time
import traceback
import sys
import xmlrpclib

import openerp

_logger = logging.getLogger(__name__)

#----------------------------------------------------------
# OpenERPSession RPC openerp backend access
#----------------------------------------------------------
class AuthenticationError(Exception):
    pass

class SessionExpiredException(Exception):
    pass

class Service(object):
    def __init__(self, session, service_name):
        self.session = session
        self.service_name = service_name

    def __getattr__(self, method):
        def proxy_method(*args):
            result = self.session.send(self.service_name, method, *args)
            return result
        return proxy_method

class Model(object):
    def __init__(self, session, model):
        self.session = session
        self.model = model
        self.proxy = self.session.proxy('object')

    def __getattr__(self, method):
        def proxy(*args, **kw):
            result = self.proxy.execute_kw(self.session._db, self.session._uid, self.session._password, self.model, method, args, kw)
            # reorder read
            if method == "read":
                if isinstance(result, list) and len(result) > 0 and "id" in result[0]:
                    index = {}
                    for r in result:
                        index[r['id']] = r
                    result = [index[x] for x in args[0] if x in index]
            return result
        return proxy

    def search_read(self, domain=None, fields=None, offset=0, limit=None, order=None, context=None):
        record_ids = self.search(domain or [], offset, limit or False, order or False, context or {})
        if not record_ids: return []
        records = self.read(record_ids, fields or [], context or {})
        return records

class OpenERPSession(object):
    """
    An OpenERP RPC session, a given user can own multiple such sessions
    in a web session.

    .. attribute:: context

        The session context, a ``dict``. Can be reloaded by calling
        :meth:`openerpweb.openerpweb.OpenERPSession.get_context`

    .. attribute:: domains_store

        A ``dict`` matching domain keys to evaluable (but non-literal) domains.

        Used to store references to non-literal domains which need to be
        round-tripped to the client browser.
    """
    def __init__(self):
        self._creation_time = time.time()
        self._db = False
        self._uid = False
        self._login = False
        self._password = False
        self._suicide = False
        self.context = {}
        self.jsonp_requests = {}     # FIXME use a LRU

    def send(self, service_name, method, *args):
        code_string = u"warning -- %s\n\n%s"
        try:
            return openerp.netsvc.dispatch_rpc(service_name, method, args)
        except openerp.osv.osv.except_osv, e:
            raise xmlrpclib.Fault(code_string % (e.name, e.value), '')
        except openerp.exceptions.Warning, e:
            raise xmlrpclib.Fault(code_string % ("Warning", e), '')
        except openerp.exceptions.AccessError, e:
            raise xmlrpclib.Fault(code_string % ("AccessError", e), '')
        except openerp.exceptions.AccessDenied, e:
            raise xmlrpclib.Fault('AccessDenied', openerp.tools.ustr(e))
        except openerp.exceptions.DeferredException, e:
            formatted_info = "".join(traceback.format_exception(*e.traceback))
            raise xmlrpclib.Fault(openerp.tools.ustr(e), formatted_info)
        except Exception, e:
            formatted_info = "".join(traceback.format_exception(*(sys.exc_info())))
            raise xmlrpclib.Fault(openerp.tools.ustr(e), formatted_info)

    def proxy(self, service):
        return Service(self, service)

    def bind(self, db, uid, login, password):
        self._db = db
        self._uid = uid
        self._login = login
        self._password = password

    def authenticate(self, db, login, password, env=None):
        uid = self.proxy('common').authenticate(db, login, password, env)
        self.bind(db, uid, login, password)

        if uid: self.get_context()
        return uid

    def assert_valid(self, force=False):
        """
        Ensures this session is valid (logged into the openerp server)
        """
        if self._uid and not force:
            return
        # TODO use authenticate instead of login
        uid = self.proxy("common").login(self._db, self._login, self._password)
        if not uid:
            raise AuthenticationError("Authentication failure")

    def ensure_valid(self):
        if self._uid:
            try:
                self.assert_valid(True)
            except Exception:
                self._uid = None

    def execute(self, model, func, *l, **d):
        self.assert_valid()
        model = self.model(model)
        r = getattr(model, func)(*l, **d)
        return r

    def exec_workflow(self, model, id, signal):
        self.assert_valid()
        r = self.proxy('object').exec_workflow(self._db, self._uid, self._password, model, signal, id)
        return r

    def model(self, model):
        """ Get an RPC proxy for the object ``model``, bound to this session.

        :param model: an OpenERP model name
        :type model: str
        :rtype: a model object
        """
        if self._db == False:
            raise SessionExpiredException("Session expired")

        return Model(self, model)

    def get_context(self):
        """ Re-initializes the current user's session context (based on
        his preferences) by calling res.users.get_context() with the old
        context

        :returns: the new context
        """
        assert self._uid, "The user needs to be logged-in to initialize his context"
        self.context = self.model('res.users').context_get() or {}
        self.context['uid'] = self._uid
        self._fix_lang(self.context)
        return self.context

    def _fix_lang(self, context):
        """ OpenERP provides languages which may not make sense and/or may not
        be understood by the web client's libraries.

        Fix those here.

        :param dict context: context to fix
        """
        lang = context['lang']

        # inane OpenERP locale
        if lang == 'ar_AR':
            lang = 'ar'

        # lang to lang_REGION (datejs only handles lang_REGION, no bare langs)
        if lang in babel.core.LOCALE_ALIASES:
            lang = babel.core.LOCALE_ALIASES[lang]

        context['lang'] = lang or 'en_US'

# vim:et:ts=4:sw=4:



# coding=utf-8
# -*- encoding: utf-8 -*-

import glob
import itertools
import json
import operator
import os

from mako.template import Template
from openerp.modules import module

from .main import module_topological_sort
from .. import http

NOMODULE_TEMPLATE = Template(u"""<!DOCTYPE html>
<html>
    <head>
        <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
        <meta http-equiv="content-type" content="text/html; charset=utf-8" />
        <title>OpenERP Testing</title>
    </head>
    <body>
        <form action="/web/tests" method="GET">
            <button name="mod" value="*">Run all tests</button>
            <ul>
            % for name, module in modules:
                <li>${name} <button name="mod" value="${module}">
                    Run Tests</button></li>
            % endfor
            </ul>
        </form>
    </body>
</html>
""", default_filters=['h'])
NOTFOUND = Template(u"""
<p>Unable to find the module [${module}], please check that the module
   name is correct and the module is on OpenERP's path.</p>
<a href="/web/tests">&lt;&lt; Back to tests</a>
""", default_filters=['h'])
TESTING = Template(u"""<!DOCTYPE html>
<html style="height: 100%">
<%def name="to_path(module, p)">/${module}/${p}</%def>
<head>
    <meta http-equiv="X-UA-Compatible" content="IE=edge"/>
    <meta http-equiv="content-type" content="text/html; charset=utf-8" />
    <title>OpenERP Web Tests</title>
    <link rel="shortcut icon" href="/web/static/src/img/favicon.ico" type="image/x-icon"/>

    <link rel="stylesheet" href="/web/static/lib/qunit/qunit.css">
    <script src="/web/static/lib/qunit/qunit.js"></script>

    <script type="text/javascript">
        var oe_db_info = ${db_info | n};
        // List of modules, each module is preceded by its dependencies
        var oe_all_dependencies = ${dependencies | n};
        QUnit.config.testTimeout = 5 * 60 * 1000;
    </script>
</head>
<body id="oe" class="openerp">
    <div id="qunit"></div>
    <div id="qunit-fixture"></div>
</body>
<!-- TODO xmo please use the regular template even for testing -->
% for module, jss, tests, templates in files:
    % for js in jss:
        % if not js.endswith('/apps.js'):
            <script src="${to_path(module, js)}"></script>
        % endif
    % endfor
    % if tests or templates:
    <script>
        openerp.testing.current_module = "${module}";
        % for template in templates:
        openerp.testing.add_template("${to_path(module, template)}");
        % endfor
    </script>
    % endif
    % if tests:
        % for test in tests:
            <script type="text/javascript" src="${to_path(module, test)}"></script>
        % endfor
    % endif
% endfor
</html>
""", default_filters=['h'])

class TestRunnerController(http.Controller):
    _cp_path = '/web/tests'

    @http.httprequest
    def index(self, req, mod=None, **kwargs):
        ms = module.get_modules()
        manifests = dict(
            (name, desc)
            for name, desc in zip(ms, map(self.load_manifest, ms))
            if desc # remove not-actually-openerp-modules
        )

        if not mod:
            return NOMODULE_TEMPLATE.render(modules=(
                (manifest['name'], name)
                for name, manifest in manifests.iteritems()
                if any(testfile.endswith('.js')
                       for testfile in manifest['test'])
            ))
        sorted_mods = module_topological_sort(dict(
            (name, manifest.get('depends', []))
            for name, manifest in manifests.iteritems()
        ))
        # to_load and to_test should be zippable lists of the same length.
        # A falsy value in to_test indicate nothing to test at that index (just
        # load the corresponding part of to_load)
        to_test = sorted_mods
        if mod != '*':
            if mod not in manifests:
                return req.not_found(NOTFOUND.render(module=mod))
            idx = sorted_mods.index(mod)
            to_test = [None] * len(sorted_mods)
            to_test[idx] = mod

        tests_candicates = [
            filter(lambda path: path.endswith('.js'),
                   manifests[mod]['test'] if mod else [])
            for mod in to_test]
        # remove trailing test-less modules
        tests = reversed(list(
            itertools.dropwhile(
                operator.not_,
                reversed(tests_candicates))))

        files = [
            (mod, manifests[mod]['js'], tests, manifests[mod]['qweb'])
            for mod, tests in itertools.izip(sorted_mods, tests)
        ]

        # if all three db_info parameters are present, send them to the page
        db_info = dict((k, v) for k, v in kwargs.iteritems()
                       if k in ['source', 'supadmin', 'password'])
        if len(db_info) != 3:
            db_info = None

        return TESTING.render(files=files, dependencies=json.dumps(
            [name for name in sorted_mods
             if module.get_module_resource(name, 'static')
             if manifests[name]['js']]), db_info=json.dumps(db_info))

    def load_manifest(self, name):
        manifest = module.load_information_from_description_file(name)
        if manifest:
            path = module.get_module_path(name)
            manifest['js'] = list(
                self.expand_patterns(path, manifest.get('js', [])))
            manifest['test'] = list(
                self.expand_patterns(path, manifest.get('test', [])))
            manifest['qweb'] = list(
                self.expand_patterns(path, manifest.get('qweb', [])))
        return manifest

    def expand_patterns(self, root, patterns):
        for pattern in patterns:
            normalized_pattern = os.path.normpath(os.path.join(root, pattern))
            for path in glob.glob(normalized_pattern):
                # replace OS path separators (from join & normpath) by URI ones
                yield path[len(root):].replace(os.path.sep, '/')



# -*- coding: utf-8 -*-
import mock
import unittest2
import openerp.addons.web.controllers.main

class TestDataSetController(unittest2.TestCase):
    def setUp(self):
        self.dataset = openerp.addons.web.controllers.main.DataSet()
        self.request = mock.Mock()
        self.read = self.request.session.model().read
        self.search = self.request.session.model().search

    def test_empty_find(self):
        self.search.return_value = []
        self.read.return_value = []

        self.assertEqual(
            self.dataset.do_search_read(self.request, 'fake.model'),
            {'records': [], 'length': 0})
        self.read.assert_called_once_with(
            [], False, self.request.context)

    def test_regular_find(self):
        self.search.return_value = [1, 2, 3]
        self.read.return_value = [
            {'id': 1, 'name': 'foo'},
            {'id': 2, 'name': 'bar'},
            {'id': 3, 'name': 'qux'}
        ]

        self.dataset.do_search_read(self.request, 'fake.model')
        self.read.assert_called_once_with(
            [1, 2, 3], False,self.request.context)

    def test_ids_shortcut(self):
        self.search.return_value = [1, 2, 3]
        self.read.return_value = [
            {'id': 1, 'name': 'foo'},
            {'id': 2, 'name': 'bar'},
            {'id': 3, 'name': 'qux'}
        ]

        self.assertEqual(
            self.dataset.do_search_read(self.request, 'fake.model', ['id']),
            {'records': [{'id': 1}, {'id': 2}, {'id': 3}], 'length': 3})
        self.assertFalse(self.read.called)



#! /usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Replacement (unofficial) `setup.py` for the `web` module.
Automatically generated by `openerpdist`. See http://noteed.com/openerpdist/.
"""

import setuptools

setuptools.setup(
    name = "openerp-web",
    version = "7.0.406",
    description = "Web",
    long_description = 
"""

OpenERP Web core module.
========================

This module provides the core of the OpenERP Web Client.
        
""",
    url = "",
    author = "",
    author_email = "TODO",
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    license = "AGPL-3",
    package_dir = {
        'openerp.addons.web': ".",
    },
    packages = ["openerp.addons.web","openerp.addons.web.controllers","openerp.addons.web.cli","openerp.addons.web.tests","openerp.addons.web.tests.qunitsuite"],
    package_data = {
        'openerp.addons.web': ["i18n/*.po*","static/test/testing.js","static/test/list-editable.js","static/test/list-utils.js","static/test/mutex.js","static/test/fulltest.html","static/test/qunit-doc.js","static/test/data.js","static/test/evals.js","static/test/list.js","static/test/form.js","static/test/Widget.js","static/test/formats.js","static/test/registry.js","static/test/class.js","static/test/rpc.js","static/test/search.js","static/test/fulltest/dataset.js","static/lib/jquery.placeholder/jquery.placeholder.js","static/lib/jquery.ba-bbq/jquery.ba-bbq.min.js","static/lib/jquery.ba-bbq/jquery.ba-bbq.js","static/lib/jquery.ui.notify/js/jquery.notify.js","static/lib/jquery.form/jquery.form.js","static/lib/jquery.ui.timepicker/js/jquery-ui-timepicker-addon.js","static/lib/jquery.autosize/jquery.autosize.js","static/lib/cleditor/jquery.cleditor.min.js","static/lib/cleditor/jquery.cleditor.js","static/lib/cleditor/images/toolbar.gif","static/lib/cleditor/images/buttons.gif","static/lib/datejs/parser.js","static/lib/datejs/date.js","static/lib/datejs/sugarpak.js","static/lib/datejs/time.js","static/lib/datejs/core.js","static/lib/datejs/extras.js","static/lib/datejs/globalization/zh-CN.js","static/lib/datejs/globalization/bg-BG.js","static/lib/datejs/globalization/en-GB.js","static/lib/datejs/globalization/smj-NO.js","static/lib/datejs/globalization/nn-NO.js","static/lib/datejs/globalization/sms-FI.js","static/lib/datejs/globalization/ar-TN.js","static/lib/datejs/globalization/es-PY.js","static/lib/datejs/globalization/zh-TW.js","static/lib/datejs/globalization/ar-EG.js","static/lib/datejs/globalization/tt-RU.js","static/lib/datejs/globalization/en-CA.js","static/lib/datejs/globalization/hu-HU.js","static/lib/datejs/globalization/es-AR.js","static/lib/datejs/globalization/de-AT.js","static/lib/datejs/globalization/sw-KE.js","static/lib/datejs/globalization/fr-CA.js","static/lib/datejs/globalization/it-CH.js","static/lib/datejs/globalization/iu-CA.js","static/lib/datejs/globalization/ar-SA.js","static/lib/datejs/globalization/fr-MC.js","static/lib/datejs/globalization/xh-ZA.js","static/lib/datejs/globalization/sr-Cyrl-CS.js","static/lib/datejs/globalization/tlh-TLH.js","static/lib/datejs/globalization/ml-IN.js","static/lib/datejs/globalization/sr-Cyrl-BA.js","static/lib/datejs/globalization/mr-IN.js","static/lib/datejs/globalization/en-PH.js","static/lib/datejs/globalization/hi-IN.js","static/lib/datejs/globalization/es-MX.js","static/lib/datejs/globalization/oc-FR.js","static/lib/datejs/globalization/ar-KW.js","static/lib/datejs/globalization/pt-PT.js","static/lib/datejs/globalization/sr-RS.js","static/lib/datejs/globalization/ky-KG.js","static/lib/datejs/globalization/fr-LU.js","static/lib/datejs/globalization/pt-BR.js","static/lib/datejs/globalization/ar-YE.js","static/lib/datejs/globalization/fi-FI.js","static/lib/datejs/globalization/eu-ES.js","static/lib/datejs/globalization/hy-AM.js","static/lib/datejs/globalization/gl-ES.js","static/lib/datejs/globalization/ab-RU.js","static/lib/datejs/globalization/fa-IR.js","static/lib/datejs/globalization/quz-BO.js","static/lib/datejs/globalization/en-BZ.js","static/lib/datejs/globalization/fr-CH.js","static/lib/datejs/globalization/ka-GE.js","static/lib/datejs/globalization/ns-ZA.js","static/lib/datejs/globalization/kn-IN.js","static/lib/datejs/globalization/kk-KZ.js","static/lib/datejs/globalization/kok-IN.js","static/lib/datejs/globalization/en-AU.js","static/lib/datejs/globalization/en-ZW.js","static/lib/datejs/globalization/is-IS.js","static/lib/datejs/globalization/de-LU.js","static/lib/datejs/globalization/hr-HR.js","static/lib/datejs/globalization/se-SE.js","static/lib/datejs/globalization/es-DO.js","static/lib/datejs/globalization/tn-ZA.js","static/lib/datejs/globalization/en-JM.js","static/lib/datejs/globalization/es-BO.js","static/lib/datejs/globalization/ar-AR.js","static/lib/datejs/globalization/zh-HK.js","static/lib/datejs/globalization/nl-BE.js","static/lib/datejs/globalization/mt-MT.js","static/lib/datejs/globalization/ro-RO.js","static/lib/datejs/globalization/ko-KP.js","static/lib/datejs/globalization/ja-JP.js","static/lib/datejs/globalization/cy-GB.js","static/lib/datejs/globalization/ar-BH.js","static/lib/datejs/globalization/es-GT.js","static/lib/datejs/globalization/zh-MO.js","static/lib/datejs/globalization/lt-LT.js","static/lib/datejs/globalization/sr@latin.js","static/lib/datejs/globalization/tr-TR.js","static/lib/datejs/globalization/syr-SY.js","static/lib/datejs/globalization/ar-AE.js","static/lib/datejs/globalization/ar-SY.js","static/lib/datejs/globalization/ms-BN.js","static/lib/datejs/globalization/dv-MV.js","static/lib/datejs/globalization/fr-FR.js","static/lib/datejs/globalization/it-IT.js","static/lib/datejs/globalization/en-US.js","static/lib/datejs/globalization/es-CO.js","static/lib/datejs/globalization/es-HN.js","static/lib/datejs/globalization/ar-LY.js","static/lib/datejs/globalization/fo-FO.js","static/lib/datejs/globalization/sk-SK.js","static/lib/datejs/globalization/es-SV.js","static/lib/datejs/globalization/sq-AL.js","static/lib/datejs/globalization/quz-EC.js","static/lib/datejs/globalization/ta-IN.js","static/lib/datejs/globalization/ar-MA.js","static/lib/datejs/globalization/uz-Cyrl-UZ.js","static/lib/datejs/globalization/en-TT.js","static/lib/datejs/globalization/ar-OM.js","static/lib/datejs/globalization/ar-QA.js","static/lib/datejs/globalization/ar-JO.js","static/lib/datejs/globalization/az-Latn-AZ.js","static/lib/datejs/globalization/sma-SE.js","static/lib/datejs/globalization/smn-FI.js","static/lib/datejs/globalization/se-FI.js","static/lib/datejs/globalization/fr-BE.js","static/lib/datejs/globalization/mi-NZ.js","static/lib/datejs/globalization/si-LK.js","static/lib/datejs/globalization/es-PE.js","static/lib/datejs/globalization/uk-UA.js","static/lib/datejs/globalization/es-CL.js","static/lib/datejs/globalization/ar-IQ.js","static/lib/datejs/globalization/sr-Latn-BA.js","static/lib/datejs/globalization/en-IE.js","static/lib/datejs/globalization/en-029.js","static/lib/datejs/globalization/en-NZ.js","static/lib/datejs/globalization/lv-LV.js","static/lib/datejs/globalization/cs-CZ.js","static/lib/datejs/globalization/vi-VN.js","static/lib/datejs/globalization/es-EC.js","static/lib/datejs/globalization/mk-MK.js","static/lib/datejs/globalization/da-DK.js","static/lib/datejs/globalization/bs-BS.js","static/lib/datejs/globalization/ar-LB.js","static/lib/datejs/globalization/sv-FI.js","static/lib/datejs/globalization/sl-SI.js","static/lib/datejs/globalization/nb-NO.js","static/lib/datejs/globalization/id-ID.js","static/lib/datejs/globalization/hr-BA.js","static/lib/datejs/globalization/az-Cyrl-AZ.js","static/lib/datejs/globalization/sa-IN.js","static/lib/datejs/globalization/th-TH.js","static/lib/datejs/globalization/ar-DZ.js","static/lib/datejs/globalization/et-EE.js","static/lib/datejs/globalization/nl-NL.js","static/lib/datejs/globalization/ms-MY.js","static/lib/datejs/globalization/quz-PE.js","static/lib/datejs/globalization/de-CH.js","static/lib/datejs/globalization/es-NI.js","static/lib/datejs/globalization/es-ES.js","static/lib/datejs/globalization/gu-IN.js","static/lib/datejs/globalization/ru-RU.js","static/lib/datejs/globalization/sr-Latn-CS.js","static/lib/datejs/globalization/de-DE.js","static/lib/datejs/globalization/sv-SE.js","static/lib/datejs/globalization/zu-ZA.js","static/lib/datejs/globalization/es-VE.js","static/lib/datejs/globalization/sma-NO.js","static/lib/datejs/globalization/es-CR.js","static/lib/datejs/globalization/be-BY.js","static/lib/datejs/globalization/bs-Latn-BA.js","static/lib/datejs/globalization/he-IL.js","static/lib/datejs/globalization/te-IN.js","static/lib/datejs/globalization/es-PA.js","static/lib/datejs/globalization/pl-PL.js","static/lib/datejs/globalization/uz-Latn-UZ.js","static/lib/datejs/globalization/zh-SG.js","static/lib/datejs/globalization/af-ZA.js","static/lib/datejs/globalization/es-UY.js","static/lib/datejs/globalization/ca-ES.js","static/lib/datejs/globalization/smj-SE.js","static/lib/datejs/globalization/el-GR.js","static/lib/datejs/globalization/pa-IN.js","static/lib/datejs/globalization/am-ET.js","static/lib/datejs/globalization/en-ZA.js","static/lib/datejs/globalization/mn-MN.js","static/lib/datejs/globalization/es-PR.js","static/lib/datejs/globalization/ur-PK.js","static/lib/datejs/globalization/ko-KR.js","static/lib/datejs/globalization/se-NO.js","static/lib/datejs/globalization/de-LI.js","static/lib/jquery.validate/jquery.validate.js","static/lib/jquery.textext/jquery.textext.js","static/lib/jquery.scrollTo/jquery.scrollTo.js","static/lib/jquery.scrollTo/jquery.scrollTo-min.js","static/lib/jquery.ui/css/smoothness/images/ui-icons_cd0a0a_256x240.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_glass_75_dadada_1x400.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_glass_55_fbf9ee_1x400.png","static/lib/jquery.ui/css/smoothness/images/ui-icons_2e83ff_256x240.png","static/lib/jquery.ui/css/smoothness/images/ui-icons_222222_256x240.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_highlight-soft_75_cccccc_1x100.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_glass_95_fef1ec_1x400.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_flat_75_ffffff_40x100.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_glass_65_ffffff_1x400.png","static/lib/jquery.ui/css/smoothness/images/ui-icons_454545_256x240.png","static/lib/jquery.ui/css/smoothness/images/ui-icons_888888_256x240.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_flat_0_aaaaaa_40x100.png","static/lib/jquery.ui/css/smoothness/images/ui-bg_glass_75_e6e6e6_1x400.png","static/lib/jquery.ui/js/jquery-ui-1.9.1.custom.js","static/lib/jquery.timeago/jquery.timeago.js","static/lib/backbone/backbone.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-alerts.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-scrollspy.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-tabs.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-modal.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-dropdown.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-buttons.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-twipsy.js","static/lib/jquery.ui.bootstrap/bootstrap/js/bootstrap-popover.js","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-Date-Range-Picker/index.html","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-Date-Range-Picker/js/daterangepicker.jQuery.js","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-Date-Range-Picker/js/daterangepicker.jQuery.compressed.js","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-Date-Range-Picker/js/date.js","static/lib/jquery.ui.bootstrap/third-party/wijmo/jquery.wijmo-open.1.5.0.min.js","static/lib/jquery.ui.bootstrap/third-party/wijmo/jquery.bgiframe-2.1.3-pre.js","static/lib/jquery.ui.bootstrap/third-party/wijmo/jquery.mousewheel.min.js","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/images/icon-media.gif","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/images/icon-zip.gif","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/images/icon-image.gif","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/images/icon-generic.gif","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/js/fileinput.jquery.js","static/lib/jquery.ui.bootstrap/third-party/jQuery-UI-FileInput/js/enhance.min.js","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_cd0a0a_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_glass_75_dadada_1x400.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_glass_55_fbf9ee_1x400.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_2e83ff_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_222222_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_highlight-soft_75_cccccc_1x100.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_inset-soft_95_fef1ec_1x100.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_glass_75_ffffff_1x400.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_glass_65_ffffff_1x400.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_454545_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_f6cf3b_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-icons_888888_256x240.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_flat_0_aaaaaa_40x100.png","static/lib/jquery.ui.bootstrap/css/custom-theme/images/ui-bg_glass_75_e6e6e6_1x400.png","static/lib/qunit/qunit.js","static/lib/jquery.MD5/jquery.md5.js","static/lib/jquery.MD5/tests/index.html","static/lib/jquery.MD5/tests/tests.js","static/lib/qweb/qweb-test-attributes.xml","static/lib/qweb/qweb.js","static/lib/qweb/qweb-benchmark.xml","static/lib/qweb/qweb-test-foreach.xml","static/lib/qweb/qweb-test-set.xml","static/lib/qweb/qweb-test-call.xml","static/lib/qweb/qweb-test-extend.xml","static/lib/qweb/qweb-test.js.html","static/lib/qweb/qweb-test-output.xml","static/lib/qweb/qweb-test-conditionals.xml","static/lib/qweb/qweb2.js","static/lib/qweb/qweb-benchmark.html","static/lib/jquery.tipsy/jquery.tipsy.js","static/lib/spinjs/spin.js","static/lib/jquery.blockUI/jquery.blockUI.js","static/lib/underscore/underscore.js","static/lib/underscore/underscore.string.js","static/lib/underscore/underscore.string.min.js","static/lib/underscore/underscore-min.js","static/lib/py.js/lib/py.js","static/lib/jquery/jquery-1.8.3.js","static/src/fixbind.js","static/src/xml/base.xml","static/src/font/mnmliconsv21-webfont.woff","static/src/font/entypo-webfont.woff","static/src/font/entypo-webfont.eot","static/src/font/mnmliconsv21-webfont.ttf","static/src/font/entypo-webfont.ttf","static/src/font/mnmliconsv21-webfont.eot","static/src/font/mnmliconsv21-webfont.svg","static/src/font/entypo-webfont.svg","static/src/js/testing.js","static/src/js/data_export.js","static/src/js/pyeval.js","static/src/js/view_list.js","static/src/js/view_tree.js","static/src/js/dates.js","static/src/js/data.js","static/src/js/view_list_editable.js","static/src/js/views.js","static/src/js/test_support.js","static/src/js/coresetup.js","static/src/js/formats.js","static/src/js/boot.js","static/src/js/chrome.js","static/src/js/corelib.js","static/src/js/search.js","static/src/js/view_form.js","static/src/img/db.png","static/src/img/stock_person.png","static/src/img/gradientlinebg.gif","static/src/img/pencil.gif","static/src/img/product.png","static/src/img/pager_last.png","static/src/img/iconset-a-help.png","static/src/img/collapse.gif","static/src/img/toggle-a-bg.png","static/src/img/partner.png","static/src/img/expand.gif","static/src/img/down-arrow.png","static/src/img/favicon.ico","static/src/img/throbber.gif","static/src/img/form_steps.png","static/src/img/back-enable.jpg","static/src/img/iconset-d-drop.gif","static/src/img/views-icons-a.png","static/src/img/pager_first.png","static/src/img/search_reset.gif","static/src/img/graph_background.png","static/src/img/logo.png","static/src/img/user_menu_avatar.png","static/src/img/logo2.png","static/src/img/form_sheetbg.png","static/src/img/view_empty_arrow.png","static/src/img/attachments-close.png","static/src/img/add-shortcut.png","static/src/img/pager_next.png","static/src/img/iconset-a-help.gif","static/src/img/remove-shortcut.png","static/src/img/box-a-header-a.gif","static/src/img/accessories-archiver.png","static/src/img/placeholder.png","static/src/img/nologo.png","static/src/img/warning.png","static/src/img/header.gif","static/src/img/pager_previous.png","static/src/img/iconset-b-remove.png","static/src/img/sep-a.gif","static/src/img/icons/STOCK_INDENT.png","static/src/img/icons/gtk-cut.png","static/src/img/icons/STOCK_MEDIA_FORWARD.png","static/src/img/icons/gtk-find-and-replace.png","static/src/img/icons/gtk-sort-descending.png","static/src/img/icons/terp-folder-violet.png","static/src/img/icons/STOCK_APPLY.png","static/src/img/icons/stock_calendar.png","static/src/img/icons/STOCK_JUSTIFY_CENTER.png","static/src/img/icons/gtk-clear.png","static/src/img/icons/STOCK_UNDELETE.png","static/src/img/icons/STOCK_GOTO_BOTTOM.png","static/src/img/icons/STOCK_EDIT.png","static/src/img/icons/STOCK_CONNECT.png","static/src/img/icons/gtk-dnd-multiple.png","static/src/img/icons/gtk-select-font.png","static/src/img/icons/STOCK_JUSTIFY_FILL.png","static/src/img/icons/gtk-print.png","static/src/img/icons/STOCK_MEDIA_REWIND.png","static/src/img/icons/STOCK_DIRECTORY_MENU.png","static/src/img/icons/terp-rating-rated.png","static/src/img/icons/terp-dolar.png","static/src/img/icons/gtk-sort-ascending.png","static/src/img/icons/terp-folder-blue.png","static/src/img/icons/gtk-media-rewind.png","static/src/img/icons/stock_person.png","static/src/img/icons/STOCK_GO_UP.png","static/src/img/icons/star-off.png","static/src/img/icons/gtk-harddisk.png","static/src/img/icons/terp-stage.png","static/src/img/icons/STOCK_HARDDISK.png","static/src/img/icons/terp-mail-replied.png","static/src/img/icons/terp-personal+.png","static/src/img/icons/gtk-save-as.png","static/src/img/icons/STOCK_FILE.png","static/src/img/icons/terp-purchase.png","static/src/img/icons/gtk-quit.png","static/src/img/icons/STOCK_MEDIA_STOP.png","static/src/img/icons/terp-dolar_ok!.png","static/src/img/icons/STOCK_SAVE.png","static/src/img/icons/terp-report.png","static/src/img/icons/gtk-new.png","static/src/img/icons/terp-document-new.png","static/src/img/icons/terp-crm.png","static/src/img/icons/gtk-go-down.png","static/src/img/icons/terp-hr.png","static/src/img/icons/gtk-goto-last.png","static/src/img/icons/gtk-about.png","static/src/img/icons/STOCK_REMOVE.png","static/src/img/icons/gtk-zoom-in.png","static/src/img/icons/gtk-goto-first.png","static/src/img/icons/STOCK_SPELL_CHECK.png","static/src/img/icons/STOCK_FULLSCREEN.png","static/src/img/icons/terp-mail-.png","static/src/img/icons/STOCK_UNDERLINE.png","static/src/img/icons/terp-stock_effects-object-colorize.png","static/src/img/icons/STOCK_DIALOG_INFO.png","static/src/img/icons/gtk-redo.png","static/src/img/icons/gtk-convert.png","static/src/img/icons/gtk-preferences.png","static/src/img/icons/STOCK_STOP.png","static/src/img/icons/STOCK_INDEX.png","static/src/img/icons/gtk-justify-left.png","static/src/img/icons/terp-emblem-important.png","static/src/img/icons/STOCK_SAVE_AS.png","static/src/img/icons/gtk-media-previous.png","static/src/img/icons/gtk-no.png","static/src/img/icons/gtk-goto-bottom.png","static/src/img/icons/gtk-ok.png","static/src/img/icons/kanban-pause.png","static/src/img/icons/terp-stock_zoom.png","static/src/img/icons/terp-account.png","static/src/img/icons/gtk-media-forward.png","static/src/img/icons/STOCK_MISSING_IMAGE.png","static/src/img/icons/terp-locked.png","static/src/img/icons/STOCK_CANCEL.png","static/src/img/icons/terp-partner.png","static/src/img/icons/STOCK_CUT.png","static/src/img/icons/terp-gtk-media-pause.png","static/src/img/icons/gtk-paste.png","static/src/img/icons/gtk-dialog-authentication.png","static/src/img/icons/STOCK_REDO.png","static/src/img/icons/STOCK_PREFERENCES.png","static/src/img/icons/gtk-connect.png","static/src/img/icons/index2.jpeg","static/src/img/icons/terp-mail_delete.png","static/src/img/icons/terp-calendar.png","static/src/img/icons/gtk-leave-fullscreen.png","static/src/img/icons/STOCK_NO.png","static/src/img/icons/STOCK_PASTE.png","static/src/img/icons/gtk-copy.png","static/src/img/icons/terp-product.png","static/src/img/icons/STOCK_MEDIA_PREVIOUS.png","static/src/img/icons/STOCK_FIND_AND_REPLACE.png","static/src/img/icons/gtk-zoom-100.png","static/src/img/icons/gtk-info.png","static/src/img/icons/STOCK_PRINT.png","static/src/img/icons/terp-mail-forward.png","static/src/img/icons/STOCK_SORT_ASCENDING.png","static/src/img/icons/STOCK_DIALOG_QUESTION.png","static/src/img/icons/index1.jpeg","static/src/img/icons/gtk-print-preview.png","static/src/img/icons/terp-graph.png","static/src/img/icons/gtk-indent.png","static/src/img/icons/STOCK_MEDIA_RECORD.png","static/src/img/icons/gtk-file.png","static/src/img/icons/terp-gtk-jump-to-rtl.png","static/src/img/icons/STOCK_MEDIA_NEXT.png","static/src/img/icons/terp-stock_format-default.png","static/src/img/icons/gtk-directory.png","static/src/img/icons/gtk-fullscreen.png","static/src/img/icons/gtk-save.png","static/src/img/icons/STOCK_PROPERTIES.png","static/src/img/icons/STOCK_GOTO_FIRST.png","static/src/img/icons/STOCK_INFO.png","static/src/img/icons/terp-personal.png","static/src/img/icons/gtk-revert-to-saved.png","static/src/img/icons/terp-folder-orange.png","static/src/img/icons/stock.png","static/src/img/icons/gtk-directory-remote.png","static/src/img/icons/STOCK_CLEAR.png","static/src/img/icons/gtk-media-next.png","static/src/img/icons/terp-stock_format-scientific.png","static/src/img/icons/STOCK_DIALOG_AUTHENTICATION.png","static/src/img/icons/terp-dialog-close.png","static/src/img/icons/gtk-network.png","static/src/img/icons/terp-mrp.png","static/src/img/icons/gtk-stop.png","static/src/img/icons/STOCK_ZOOM_OUT.png","static/src/img/icons/gtk-go-up.png","static/src/img/icons/STOCK_GO_DOWN.png","static/src/img/icons/terp-marketing.png","static/src/img/icons/STOCK_DISCONNECT.png","static/src/img/icons/STOCK_OK.png","static/src/img/icons/terp-check.png","static/src/img/icons/gtk-zoom-fit.png","static/src/img/icons/STOCK_CONVERT.png","static/src/img/icons/stock_translate.png","static/src/img/icons/gtk-justify-right.png","static/src/img/icons/STOCK_SELECT_FONT.png","static/src/img/icons/STOCK_ITALIC.png","static/src/img/icons/gtk-delete.png","static/src/img/icons/terp-tools.png","static/src/img/icons/gtk-add.png","static/src/img/icons/STOCK_COPY.png","static/src/img/icons/STOCK_ADD.png","static/src/img/icons/gtk-media-record.png","static/src/img/icons/gtk-close.png","static/src/img/icons/gtk-media-stop.png","static/src/img/icons/save-document.png","static/src/img/icons/terp-translate.png","static/src/img/icons/gtk-underline.png","static/src/img/icons/kanban-stop.png","static/src/img/icons/terp-mail-message-new.png","static/src/img/icons/STOCK_GO_BACK.png","static/src/img/icons/STOCK_GO_FORWARD.png","static/src/img/icons/gtk-edit.png","static/src/img/icons/gtk-apply.png","static/src/img/icons/STOCK_PRINT_PREVIEW.png","static/src/img/icons/gtk-cdrom.png","static/src/img/icons/color-picker.png","static/src/img/icons/terp-call-start.png","static/src/img/icons/STOCK_COLOR_PICKER.png","static/src/img/icons/STOCK_GOTO_TOP.png","static/src/img/icons/STOCK_CDROM.png","static/src/img/icons/terp-gtk-select-all.png","static/src/img/icons/terp-face-plain.png","static/src/img/icons/STOCK_MEDIA_PAUSE.png","static/src/img/icons/gtk-help.png","static/src/img/icons/gtk-open.png","static/src/img/icons/gtk-dialog-info.png","static/src/img/icons/terp-go-month.png","static/src/img/icons/gtk-color-picker.png","static/src/img/icons/STOCK_ZOOM_FIT.png","static/src/img/icons/terp-stock_align_left_24.png","static/src/img/icons/STOCK_ZOOM_IN.png","static/src/img/icons/terp-personal-.png","static/src/img/icons/terp-gdu-smart-failing.png","static/src/img/icons/gtk-unindent.png","static/src/img/icons/terp-gtk-go-back-ltr.png","static/src/img/icons/terp-go-home.png","static/src/img/icons/stock_log.png","static/src/img/icons/terp-sale.png","static/src/img/icons/STOCK_GOTO_LAST.png","static/src/img/icons/terp-accessories-archiver.png","static/src/img/icons/STOCK_NETWORK.png","static/src/img/icons/terp-project.png","static/src/img/icons/STOCK_HOME.png","static/src/img/icons/terp-folder-yellow.png","static/src/img/icons/gtk-italic.png","static/src/img/icons/STOCK_UNDO.png","static/src/img/icons/gtk-go-back.png","static/src/img/icons/STOCK_MEDIA_PLAY.png","static/src/img/icons/gtk-select-color.png","static/src/img/icons/gtk-zoom-out.png","static/src/img/icons/gtk-home.png","static/src/img/icons/STOCK_STRIKETHROUGH.png","static/src/img/icons/gtk-find.png","static/src/img/icons/gtk-refresh.png","static/src/img/icons/gtk-dialog-warning.png","static/src/img/icons/terp-gtk-jump-to-ltr.png","static/src/img/icons/gtk-media-pause.png","static/src/img/icons/STOCK_BOLD.png","static/src/img/icons/gtk-justify-center.png","static/src/img/icons/terp-idea.png","static/src/img/icons/STOCK_CLOSE.png","static/src/img/icons/gtk-floppy.png","static/src/img/icons/terp-gnome-cpu-frequency-applet+.png","static/src/img/icons/terp-accessories-archiver-minus.png","static/src/img/icons/STOCK_SELECT_COLOR.png","static/src/img/icons/STOCK_REVERT_TO_SAVED.png","static/src/img/icons/gtk-properties.png","static/src/img/icons/gtk-bold.png","static/src/img/icons/STOCK_DIALOG_WARNING.png","static/src/img/icons/STOCK_EXECUTE.png","static/src/img/icons/STOCK_DND_MULTIPLE.png","static/src/img/icons/STOCK_QUIT.png","static/src/img/icons/terp-camera_test.png","static/src/img/icons/STOCK_FLOPPY.png","static/src/img/icons/STOCK_DIRECTORY.png","static/src/img/icons/STOCK_DND.png","static/src/img/icons/STOCK_HELP.png","static/src/img/icons/gtk-cancel.png","static/src/img/icons/STOCK_NEW.png","static/src/img/icons/gtk-goto-top.png","static/src/img/icons/STOCK_YES.png","static/src/img/icons/STOCK_JUSTIFY_LEFT.png","static/src/img/icons/terp-accessories-archiver+.png","static/src/img/icons/STOCK_DELETE.png","static/src/img/icons/STOCK_LEAVE_FULLSCREEN.png","static/src/img/icons/gtk-execute.png","static/src/img/icons/kanban-apply.png","static/src/img/icons/STOCK_UNINDENT.png","static/src/img/icons/stock_graph.png","static/src/img/icons/gtk-paste-v.png","static/src/img/icons/terp-go-today.png","static/src/img/icons/terp-folder-green.png","static/src/img/icons/gtk-dialog-error.png","static/src/img/icons/gtk-undelete.png","static/src/img/icons/terp-stock.png","static/src/img/icons/gtk-spell-check.png","static/src/img/icons/gtk-dialog-question.png","static/src/img/icons/gtk-justify-fill.png","static/src/img/icons/terp-administration.png","static/src/img/icons/gtk-dnd.png","static/src/img/icons/STOCK_ABOUT.png","static/src/img/icons/terp-gtk-stop.png","static/src/img/icons/STOCK_REFRESH.png","static/src/img/icons/terp-go-year.png","static/src/img/icons/gtk-jump-to.png","static/src/img/icons/gtk-index.png","static/src/img/icons/gtk-disconnect.png","static/src/img/icons/terp-gtk-go-back-rtl.png","static/src/img/icons/STOCK_JUSTIFY_RIGHT.png","static/src/img/icons/STOCK_OPEN.png","static/src/img/icons/gtk-remove.png","static/src/img/icons/gtk-go-forward.png","static/src/img/icons/gtk-undo.png","static/src/img/icons/STOCK_FIND.png","static/src/img/icons/gtk-strikethrough.png","static/src/img/icons/star-on.png","static/src/img/icons/gtk-media-play.png","static/src/img/icons/STOCK_DIALOG_ERROR.png","static/src/img/icons/gtk-missing-image.png","static/src/img/icons/terp-go-week.png","static/src/img/icons/STOCK_JUMP_TO.png","static/src/img/icons/STOCK_ZOOM_100.png","static/src/img/icons/gtk-yes.png","static/src/img/icons/stock_gantt.png","static/src/img/icons/STOCK_SORT_DESCENDING.png","static/src/img/icons/terp-stock_symbol-selection.png","static/src/img/ui/group-expanded.png","static/src/img/ui/field_calendar.png","static/src/img/ui/group-folded.png","i18n/sk.po","i18n/fa.po","i18n/id.po","i18n/hr.po","i18n/da.po","i18n/sl.po","i18n/es_MX.po","i18n/fr_CA.po","i18n/ru.po","i18n/zh_TW.po","i18n/en_GB.po","i18n/cs.po","i18n/sq.po","i18n/fi.po","i18n/bn.po","i18n/pt_BR.po","i18n/ko.po","i18n/gu.po","i18n/bg.po","i18n/es_PE.po","i18n/ja.po","i18n/ro.po","i18n/hi.po","i18n/sr@latin.po","i18n/es.po","i18n/ka.po","i18n/mn.po","i18n/es_CL.po","i18n/it.po","i18n/nl.po","i18n/sv.po","i18n/es_EC.po","i18n/uk.po","i18n/bs.po","i18n/tr.po","i18n/en_AU.po","i18n/de.po","i18n/et.po","i18n/lo.po","i18n/lt.po","i18n/hu.po","i18n/vi.po","i18n/eu.po","i18n/gl.po","i18n/lv.po","i18n/mk.po","i18n/ca.po","i18n/fr.po","i18n/zh_CN.po","i18n/pl.po","i18n/he.po","i18n/th.po","i18n/nl_BE.po","i18n/es_DO.po","i18n/nb.po","i18n/ar.po","i18n/pt.po","i18n/es_CR.po","doc/images/tests.png","doc/images/db-query.png","doc/images/runner2.png","doc/images/runner.png","doc/images/tests3.png","doc/images/tests2.png","doc/_templates/sidebarlogo.html","doc/_templates/sidebarintro.html","doc/module/web_example.xml","doc/module/static/src/xml/web_example.xml","doc/module/static/src/js/first_module.js","doc/_themes/flask_small/layout.html","doc/_themes/flask/relations.html","doc/_themes/flask/layout.html","doc/_static/openerp.png","tests/qunitsuite/grunt/bootstrap.js","tests/qunitsuite/grunt/qunit-phantomjs-bridge.js","static/src/xml/*.xml","static/lib/jquery.ui.bootstrap/css/custom-theme/jquery-ui-1.9.0.custom.css","static/lib/jquery.ui.timepicker/css/jquery-ui-timepicker-addon.css","static/lib/jquery.ui.notify/css/ui.notify.css","static/lib/jquery.tipsy/tipsy.css","static/lib/jquery.textext/jquery.textext.css","static/src/css/base.css","static/src/css/data_export.css","static/lib/cleditor/jquery.cleditor.css","static/src/fixbind.js","static/lib/datejs/globalization/en-US.js","static/lib/datejs/core.js","static/lib/datejs/parser.js","static/lib/datejs/sugarpak.js","static/lib/datejs/extras.js","static/lib/jquery/jquery-1.8.3.js","static/lib/jquery.MD5/jquery.md5.js","static/lib/jquery.form/jquery.form.js","static/lib/jquery.validate/jquery.validate.js","static/lib/jquery.ba-bbq/jquery.ba-bbq.js","static/lib/spinjs/spin.js","static/lib/jquery.autosize/jquery.autosize.js","static/lib/jquery.blockUI/jquery.blockUI.js","static/lib/jquery.placeholder/jquery.placeholder.js","static/lib/jquery.ui/js/jquery-ui-1.9.1.custom.js","static/lib/jquery.ui.timepicker/js/jquery-ui-timepicker-addon.js","static/lib/jquery.ui.notify/js/jquery.notify.js","static/lib/jquery.deferred-queue/jquery.deferred-queue.js","static/lib/jquery.scrollTo/jquery.scrollTo-min.js","static/lib/jquery.tipsy/jquery.tipsy.js","static/lib/jquery.textext/jquery.textext.js","static/lib/jquery.timeago/jquery.timeago.js","static/lib/qweb/qweb2.js","static/lib/underscore/underscore.js","static/lib/underscore/underscore.string.js","static/lib/backbone/backbone.js","static/lib/cleditor/jquery.cleditor.js","static/lib/py.js/lib/py.js","static/src/js/boot.js","static/src/js/testing.js","static/src/js/pyeval.js","static/src/js/corelib.js","static/src/js/coresetup.js","static/src/js/dates.js","static/src/js/formats.js","static/src/js/chrome.js","static/src/js/views.js","static/src/js/data.js","static/src/js/data_export.js","static/src/js/search.js","static/src/js/view_form.js","static/src/js/view_list.js","static/src/js/view_list_editable.js","static/src/js/view_tree.js"],
    },
    install_requires = ["mock"],
    tests_require = ["unittest2"],
)



import logging
import optparse
import sys

import unittest2

import openerp
import openerp.addons.web.tests

_logger = logging.getLogger(__name__)

class TestJs(openerp.cli.Command):
    def run(self, args):
        self.parser = parser = optparse.OptionParser()
        parser.add_option("-d", "--database", dest="db_name", default=False, help="specify the database name")
        parser.add_option("--xmlrpc-port", dest="xmlrpc_port", default=8069, help="specify the TCP port for the XML-RPC protocol", type="int")
        # proably need to add both --superadmin-password and --database-admin-password
        self.parser.parse_args(args)

        # test ony uses db_name xmlrpc_port admin_passwd, so use the server one for the actual parsing

        config = openerp.tools.config
        config.parse_config(args)
        # needed until runbot is fixed
        config['db_password'] = config['admin_passwd']

        # run js tests
        openerp.netsvc.init_alternative_logger()
        suite = unittest2.TestSuite()
        suite.addTests(unittest2.TestLoader().loadTestsFromModule(openerp.addons.web.tests.test_js))
        r = unittest2.TextTestRunner(verbosity=2).run(suite)
        if r.errors or r.failures:
            sys.exit(1)

# vim:et:ts=4:sw=4:
